from math import *
import numpy as np

from .loadnavdata import load_navdata
from bluesky.tools import geo
from bluesky.tools.aero import nm
from bluesky.tools.misc import findall
import bluesky as bs

class Navdatabase:
    """
    Navdatabase class definition : command stack & processing class

    Methods:
        Navdatabase()          :  constructor

        findid(txt,lat,lon)    : find a nav closest to lat,lon


    Members:
        wpid                      : list of identifier/short names
        wpname                    : long name
        wptype                    : type of waypoint (yet unused)
        wplat                     : latitude
        wplon                     : longitude
        wpco                      : country code

        apid                      : list of identifier/short names
        apname                    : long name
        aplat                     : latitude
        aplon                     : longitude
        aptype                    : type of airport (1=large, 2=medium, 3=small)
        apmaxrwy                  : max rwy length in meters
        apco                      : country code
        apelev                    : country code


    Created by  : Jacco M. Hoekstra (TU Delft)
    """

    def __init__(self):
        """The navigation database: Contains waypoint, airport, airway, and sector data, but also
           geographical graphics data."""
        # Variables are initialized in reset()
        self.reset()

    def reset(self):
        print("Loading global navigation database...")
        wptdata, aptdata, awydata, firdata, codata, rwythresholds = load_navdata()

        # Get waypoint data
        self.wpid     = wptdata['wpid']       # identifier (string)
        self.wplat    = wptdata['wplat']      # latitude [deg]
        self.wplon    = wptdata['wplon']      # longitude [deg]
        self.wptype   = wptdata['wptype']     # type (string)
        self.wpelev   = wptdata['wpelev']     # elevation [m]
        self.wpvar    = wptdata['wpvar']      # magn variation [deg]
        self.wpfreq   = wptdata['wpfreq']       # frequency [kHz/MHz]
        self.wpdesc   = wptdata['wpdesc']     # description

        # Get airway legs data
        self.awfromwpid = awydata['awfromwpid']  # identifier (string)
        self.awfromlat  = awydata['awfromlat']   # latitude [deg]
        self.awfromlon  = awydata['awfromlon']   # longitude [deg]
        self.awtowpid   = awydata['awtowpid']    # identifier (string)
        self.awtolat    = awydata['awtolat']     # latitude [deg]
        self.awtolon    = awydata['awtolon']     # longitude [deg]
        self.awid       = awydata['awid']        # airway identifier (string)
        self.awndir     = awydata['awndir']      # number of directions (1 or 2)
        self.awlowfl    = awydata['awlowfl']     # lower flight level (int)
        self.awupfl     = awydata['awupfl']      # upper flight level (int)

        # Get airpoint data
        self.aptid     = aptdata['apid']      # 4 char identifier (string)
        self.aptname   = aptdata['apname']    # full name
        self.aptlat    = aptdata['aplat']     # latitude [deg]
        self.aptlon    = aptdata['aplon']     # longitude [deg]
        self.aptmaxrwy = aptdata['apmaxrwy']  # max runway length [m]
        self.aptype    = aptdata['aptype']    # type (int, 1=large, 2=medium, 3=small)
        self.aptco     = aptdata['apco']      # two char country code (string)
        self.aptelev   = aptdata['apelev']    # field elevation in meters [m] above mean sea level

        # Get FIR data
        self.fir      = firdata['fir']        # fir name
        self.firlat0  = firdata['firlat0']    # start lat of a line of border
        self.firlon0  = firdata['firlon0']    # start lon of a line of border
        self.firlat1  = firdata['firlat1']    # end lat of a line of border
        self.firlon1  = firdata['firlon1']    # end lon of a line of border

        # Get country code data
        self.coname   = codata['coname']      # country full name
        self.cocode2  = codata['cocode2']     # country code A2 (asscii2) 2 chars
        self.cocode3  = codata['cocode3']     # country code A3 (asscii2) 3 chars
        self.conr     = codata['conr']        # country icao number

        self.rwythresholds = rwythresholds

    def defwpt(self,name=None,lat=None,lon=None,wptype=None):
        # Prevent polluting the database: check arguments
        if name is None or name == "":
            return False, "Insufficient arguments"
        elif name.isdigit():
            return False, "Name needs to start with an alphabetical character"

        # DEL command: give info on waypoint (shudl work wit or without lat,lon, may be clicked by accident
        elif (not wptype==None and (wptype.upper()=="DEL" or wptype.upper() =="DELETE")) or \
            (type(lon)==str and (lon.upper()=="DEL" or lon.upper=="DELETE")):
            return self.delwpt(name)


        # No data: give info on waypoint
        elif lat==None or lon==None:
            if self.wpid.count(name.upper()) > 0:
                i = self.getwpidx(name.upper(), bs.ref.lat, bs.ref.lon)
                txt = f"{self.wpid[i]} : {self.wplat[i]}, {self.wplon[i]}"
                if len(self.wptype[i]) > 0:
                    txt = txt + "  " + self.wptype[i]
                return True, txt

            # Waypoint name is free
            else:
                return True, f"Waypoint {name.upper()} does not yet exist."

        # Still here? So there is data, then we add this waypoint
        self.wpid.append(name.upper())
        self.wplat = np.append(self.wplat,lat)
        self.wplon = np.append(self.wplon,lon)

        if wptype == None:
            self.wptype.append("")
        else:
            self.wptype.append(wptype)

        self.wpelev.append(0.0)               # elevation [m]
        self.wpvar.append(0.0)                # magn variation [deg]
        self.wpfreq.append(0.0)               # frequency [kHz/MHz]
        self.wpdesc.append("Custom waypoint") # description

         # Update screen info
        bs.scr.addnavwpt(name.upper(),lat,lon)

        return True #,name.upper()+" added to navdb."

    def delwpt(self,name=''):
        """ Delete a waypoint"""
        if self.wpid.count(name.upper()) <= 0:
            return False,"Waypoint " + name.upper() + " does not exist."

        idx = len(self.wpid) - self.wpid[::-1].index(name) - 1 # Search from back of list

        del self.wpid[idx]   # wp name

        np.delete(self.wplat,idx)  # wp lat
        np.delete (self.wplon,idx) # wp lon

        del self.wptype[idx]        # Waypoint type
        del self.wpelev[idx]        # elevation [m]
        del self.wpvar[idx]         # magn variation [deg]
        del self.wpfreq[idx]        # frequency [kHz/MHz]
        del self.wpdesc[idx]        # description

         # Update screen info (delete necessary there?)
        bs.scr.removenavwpt(name.upper())

        return True #,name.upper()+" deleted from navdb."

    def getwpidx(self, txt, reflat: float|None = None, reflon: float|None = None):
        """Get waypoint index to access data"""
        name = txt.upper()
        try:
            i = self.wpid.index(name)
        except:
            return -1

        # if no pos is specified, get first occurence
        if reflat is None:
            return i

        # If pos is specified check for more and return closest
        else:
            idx = []
            idx.append(i)
            found = True
            while i < len(self.wpid) - 1 and found:
                try:
                    i = self.wpid.index(name, i + 1)
                    idx.append(i)
                except:
                    found = False
            if len(idx) == 1:
                return idx[0]
            else:
                imin = idx[0]
                dmin = geo.kwikdist(reflat, reflon, self.wplat[imin], self.wplon[imin])
                for i in idx[1:]:
                    d = geo.kwikdist(reflat, reflon, self.wplat[i], self.wplon[i])
                    if d < dmin:
                        imin = i
                        dmin = d
                return imin

    def getwpindices(self, txt, reflat: float|None = None, reflon: float|None = None, crit=1852.0):
        """Get waypoint index to access data"""
        name = txt.upper()
        try:
            i = self.wpid.index(name)
        except:
            return [-1]

        # if no pos is specified, get first occurence
        if reflat is None:
            return [i]

        # If pos is specified check for more and return closest
        else:
            idx = findall(self.wpid,name) # find indices of al occurences

            if len(idx) == 1:
                return [idx[0]]
            else:
                imin = idx[0]
                dmin = geo.kwikdist(reflat, reflon, self.wplat[imin], self.wplon[imin])
                for i in idx[1:]:
                    d = geo.kwikdist(reflat, reflon, self.wplat[i], self.wplon[i])
                    if d < dmin:
                        imin = i
                        dmin = d
                # Find co-located
                indices = [imin]
                for i in idx:
                    if i!=imin:
                        dist = nm*geo.kwikdist(self.wplat[i], self.wplon[i], \
                                            self.wplat[imin], self.wplon[imin])
                        if dist<=crit:
                            indices.append(i)

                return indices

    def getaptidx(self, txt):
        """Get waypoint index to access data"""
        try:
            return self.aptid.index(txt.upper())
        except:
            return -1

    def getinear(self, wlat, wlon, lat, lon):  # lat,lon in degrees
        # t0 = time.clock()
        f = cos(radians(lat))
        dlat = (wlat - lat + 180.) % 360. - 180.
        dlon = f * ((wlon - lon + 180.) % 360. - 180.)
        d2 = dlat * dlat + dlon * dlon
        idx = np.argmin(d2)
        # dt = time.clock()-t0
        # print dt
        return idx

    def getwpinear(self, lat, lon):  # lat,lon in degrees
        """Get closest waypoint index"""
        return self.getinear(self.wplat, self.wplon, lat, lon)

    def getapinear(self, lat, lon):  # lat,lon in degrees
        """Get closest airport index"""
        return self.getinear(self.aptlat, self.aptlon, lat, lon)

    def getinside(self, wlat, wlon, lat0, lat1, lon0, lon1):
        """Get indices inside given box"""
        # t0 = time.clock()
        if lat0 < lat1:
            arr = np.where((wlat > lat0) * (wlat < lat1) * (wlon > lon0) * (wlon < lon1))
        else:
            arr = np.where((wlat > lat1) + (wlat < lat0) * (wlon > lon0) * (wlon < lon1))

        # dt = time.clock()-t0
        # print dt
        return list(arr[0])  # Get indices

    def getwpinside(self, lat0, lat1, lon0, lon1):
        """Get waypoint indices inside box"""
        return self.getinside(self.wplat, self.wplon, lat0, lat1, lon0, lon1)

    def getapinside(self, lat0, lat1, lon0, lon1):
        """Get airport indicex inside box"""
        return self.getinside(self.aptlat, self.aptlon, lat0, lat1, lon0, lon1)

    # returns all runways of given airport
    def listairway(self, airwayid):
        awkey = airwayid.upper()

        airway = []     # identifier of waypoint   0 .. N-1

        # Does this airway exist?
        if self.awid.count(awkey)>0:
            # Collect leg indices
            i = 0
            found = True
            legs  = []  # Alle leg incl. duplicate legs
            left  = []  # wps in left column in file
            right = []  # wps in right coumn in file

            idx = findall(self.awid,awkey)
            for i in idx:
                newleg = self.awfromwpid[i]+"-"+self.awtowpid[i]
                if newleg not in legs:
                    legs.append(newleg)
                    left.append(self.awfromwpid[i])
                    right.append(self.awtowpid[i])

            # Not found: return
            if len(legs)==0:
                return []

            # Count wps to see when we have all segments
            unused =  len(left)+len(right)

            while unused>0 and not left==len(left)*[""]:

                # Find start of a segment
                wps  = left+right
                iwps = 0
                while iwps<len(wps) and wps.count(wps[iwps])>1:
                    iwps = iwps + 1

                i = iwps%len(left)
                j = int(iwps/len(left))

                # Catch single lost wps
                if j>1 or iwps>len(wps):
                    break

                # Sort
                wps = [left,right]
                segment = []

                segready = False
                while not segready:

                    # Get leg
                    curwp  = wps[j][i]
                    nextwp = wps[1-j][i]

                    # Update admin of to do wplist
                    unused      = unused - 2
                    wps[j][i]   = ""
                    wps[1-j][i] = ""

                    # Add first wp to segment
                    segment.append(curwp)

                    # Find next lef with nextwp
                    if wps[0].count(nextwp)>0:
                        j  = 0
                        i  = wps[0].index(nextwp)
                        found = True

                    elif wps[1].count(nextwp)>0:
                        i  = wps[1].index(nextwp)
                        j  = 1
                        found = True
                    else:
                        found = False

                    # This segemnt done?
                    segready = (not found) or curwp=="" or nextwp==""

                # Also add final nextwp of this segment
                segment.append(nextwp)

                # Airway cab have multiple separate segments
                airway.append(segment)

                # Ready for next segment
                left  = wps[0]
                right = wps[1]


        return airway #,connect

    def listconnections(self, wpid,wplat,wplon):

        # Return list of connecting airway legs
        connect = []

        # Check from-list first
        if wpid in self.awfromwpid:
            idx = findall(self.awfromwpid,wpid)
            for i in idx:
                newitem = [self.awid[i],self.awtowpid[i]]
                if (newitem not in connect) and \
                         geo.kwikdist(self.awfromlat[i],self.awfromlon[i],
                                  wplat,wplon) < 10.:
                    connect.append(newitem)

        # Check to-list nextt
        if wpid in self.awtowpid:
            idx = findall(self.awtowpid,wpid)
            for i in idx:
                newitem = [self.awid[i],self.awfromwpid[i]]
                if (newitem not in connect) and \
                         geo.kwikdist(self.awtolat[i],self.awtolon[i],
                                  wplat,wplon) < 10.:
                    connect.append(newitem)

        return connect # return list of [awid,wpid]



""" Implementation of BlueSky's plugin system. """
import ast
import importlib
from pathlib import Path
import bluesky as bs
from bluesky import plugins
from bluesky import settings
from bluesky.core import timed_function, varexplorer as ve
from bluesky import stack

# Register settings defaults
settings.set_variable_defaults(plugin_path='plugins', enabled_plugins=['datafeed'])


class Plugin:
    ''' BlueSky plugin class.
        This class is used internally to store information about bluesky
        plugins that were found in the search directory. '''

    # Dictionary of all available plugins
    plugins = dict()

    # Plugins not meant for this process type (i.e., sim/gui)
    plugins_ext = list()

    loaded_plugins = dict()

    def __init__(self, fullname):
        self.fullname = fullname

        self.plugin_doc   = ''
        self.plugin_name  = ''
        self.plugin_type  = ''
        self.plugin_stack = []
        self.loaded = False
        self.imp = None

    def _load(self):
        ''' Load this plugin. '''
        if self.loaded:
            return False, f'Plugin {self.plugin_name} already loaded'

        try:
            # Load the plugin, or get already loaded module from sys modules
            self.imp = importlib.import_module(self.fullname)

            # Initialize the plugin
            result = self.imp.init_plugin()
            config = result if isinstance(result, dict) else result[0]

            if self.plugin_type == 'sim':
                dt = max(config.get('update_interval', 0.0), bs.sim.simdt)
                # Add timed functions if present
                for hook in ('preupdate', 'update', 'reset'):
                    fun = config.get(hook)
                    if fun:
                        timed_function(
                            fun, name=f'{self.plugin_name}.{fun.__name__}', dt=dt, hook=hook)

                # Add the plugin as data parent to the variable explorer
                ve.register_data_parent(self.imp, self.plugin_name.lower())

            if isinstance(result, (tuple, list)) and len(result) > 1:
                stackfuns = result[1]
                # Add the plugin's stack functions to the stack
                bs.stack.append_commands(stackfuns)
            

            return True, 'Successfully loaded plugin %s' % self.plugin_name
        except ImportError as e:
            print('BlueSky plugin system failed to load', self.plugin_name, ':', e)
            return False, f'Failed to load {self.plugin_name}'

    @classmethod
    def load(cls, name):
        ''' Load a plugin by name. '''
        plugin = cls.plugins.get(name)
        if plugin is None:
            return False, f'Error loading plugin: plugin {name} not found.'

        # Try to load the plugin
        success, msg = plugin._load()
        if success:
            cls.loaded_plugins[name] = plugin
        return success, msg

    @classmethod
    def find_plugins(cls, reqtype):
        ''' Create plugin wrapper objects based on source code of potential plug-in files. '''
        for path in (Path(p) for p in plugins.__spec__.submodule_search_locations):
            for fname in path.glob('**/*.py'):
                submod = fname.relative_to(path).parent.as_posix().replace('/', '.')
                fullname = f'bluesky.plugins.{fname.stem}' if submod == '.' else \
                           f'bluesky.plugins.{submod}.{fname.stem}'
                with open(fname, 'rb') as f:
                    source = f.read()
                    try:
                        tree = ast.parse(source)
                    except:
                        # Failed to parse source code, continue to next file
                        continue

                    ret_dicts = []
                    ret_names = ['', '']
                    for item in tree.body:
                        if isinstance(item, ast.FunctionDef) and item.name == 'init_plugin':
                            for iitem in reversed(item.body):
                                # Return value of init_plugin should always be a tuple of two dicts
                                # The first dict is the plugin config dict, the second dict is the stack function dict
                                if isinstance(iitem, ast.Return):
                                    if isinstance(iitem.value, ast.Tuple):
                                        ret_dicts = iitem.value.elts
                                    else:
                                        ret_dicts = [iitem.value]
                                    if len(ret_dicts) not in (1, 2):
                                        print(f"{fname} looks like a plugin, but init_plugin() doesn't return one or two dicts")
                                        continue
                                    ret_names = [el.id if isinstance(el, ast.Name) else '' for el in ret_dicts]

                                # Check if this is the assignment of one of the return values
                                if isinstance(iitem, ast.Assign) and isinstance(iitem.value, ast.Dict):
                                    for i, name in enumerate(ret_names):
                                        if iitem.targets[0].id == name:
                                            ret_dicts[i] = iitem.value

                            # Parse the config dict
                            cfgdict = {k.s:v for k,v in zip(ret_dicts[0].keys, ret_dicts[0].values)}
                            plugintype = cfgdict.get('plugin_type')
                            if plugintype is None:
                                print(f'{fname} looks like a plugin, but no plugin type (sim/gui) is specified. ' 
                                        'To fix this, add the element plugin_type to the configuration dictionary that is returned from init_plugin()')
                                continue
                            if plugintype.s == reqtype:
                                # This is the initialization function of a bluesky plugin. Parse the contents
                                plugin = Plugin(fullname)
                                plugin.plugin_doc = ast.get_docstring(tree)
                                plugin.plugin_name = cfgdict['plugin_name'].s
                                plugin.plugin_type = cfgdict['plugin_type'].s

                                # Parse the stack function dict
                                if len(ret_dicts) > 1:
                                    stack_keys       = [el.s for el in ret_dicts[1].keys]
                                    stack_docs       = [el.elts[-1].s for el in ret_dicts[1].values]
                                    plugin.plugin_stack = list(zip(stack_keys, stack_docs))
                                # Add plugin to the dict of available plugins
                                cls.plugins[plugin.plugin_name.upper()] = plugin
                            else:
                                cls.plugins_ext.append(cfgdict['plugin_name'].s.upper())


def init(mode):
    ''' Initialization function of the plugin system.'''
    # Set plugin type for this instance of BlueSky
    req_type = 'sim' if mode[:3] == 'sim' else 'gui'
    oth_type = 'gui' if mode[:3] == 'sim' else 'sim'

    # Find available plugins
    Plugin.find_plugins(req_type)
    # Load plugins selected in config
    for pname in settings.enabled_plugins:
        if pname.upper() not in Plugin.plugins_ext:
            success = Plugin.load(pname.upper())
            print(success[1])

    # Create the plugin management stack command
    @stack.command(name='PLUGINS', aliases=('PLUGIN', 'PLUG-IN', 'PLUG-INS', f'{req_type.upper()}PLUGIN'))
    def manage(cmd: 'txt' = 'LIST', plugin_name: 'txt' = ''):
        ''' List all plugins, load a plugin, or remove a loaded plugin.'''
        if cmd == 'LIST':
            running = set(Plugin.loaded_plugins.keys())
            available = set(Plugin.plugins.keys()) - running
            text = f'\nCurrently running {req_type} plugins: {", ".join(running)}'
            if available:
                text += f'\nAvailable {req_type} plugins: {", ".join(available)}'
            else:
                text += f'\nNo additional {req_type} plugins available.'
            # Also let other side print the list of plugins
            stack.forward()
            return True, text

        if cmd in ('LOAD', 'ENABLE') or not plugin_name:
            # If no command is given, assume user tries to load a plugin
            success, msg = Plugin.load(plugin_name or cmd)
            # If we're the first and plugin is not found here, send it on
            if not success:
                stack.forward()
                return True
            return success, msg

        return False, f'Unknown command {cmd}'



'''
    Provides BlueSky derived classes in BlueSky with the following functionality:
    - allow replaceable implementations (in plugins) that can be selected at runtime.
    - allow class and instance methods to be decorated as
      - stack functions
      - network subscribers
      - shared state subscribers
      - timed functions
      - signal subscribers
'''
import inspect
from typing import Dict, Optional, Type, ClassVar

from bluesky.core.funcobject import FuncObject

# Global dictionary of replaceable BlueSky classes
replaceables: Dict[str, Type['Base']] = dict()


def reset():
    ''' Reset all replaceables to their default implementation. '''
    for base in replaceables.values():
        base.selectdefault()


def select_implementation(basename='', implname=''):
    ''' Stack function to select an implementation for the construction of
        objects of the class corresponding to basename. '''
    if not basename:
        return True, 'Replaceable classes in Bluesky:\n' + \
            ', '.join(replaceables)
    base = replaceables.get(basename.upper(), None)
    if not base:
        return False, f'Replaceable {basename} not found.'
    impls = base.derived()
    if not implname:
        return True, f'Current implementation for {basename}: {base._generator.__name__}\n' + \
            f'Available implementations for {basename}:\n' + \
            ', '.join(impls)

    impl = impls.get(base.__name__ if implname == 'BASE' else implname)
    if not impl:
        return False, f'Implementation {implname} not found for replaceable {basename}.'
    impl.select()
    return True, f'Selected implementation {implname} for replaceable {basename}'


class Base:
    '''
        Base: Provides BlueSky derived classes in BlueSky with the following functionality:
        - allow replaceable implementations (in plugins) that can be selected at runtime.
        - allow class and instance methods to be decorated as
          - stack functions
          - network subscribers
          - shared state subscribers
          - timed functions
          - signal subscribers
    '''
    # Class variables that will be set for subclasses
    _baseimpl: ClassVar[Type['Base']]
    _default: ClassVar[str]
    _generator: ClassVar[Type['Base']]
    _selinstance: ClassVar[Optional['Base']]
    __func_objects__: ClassVar[Dict[str, FuncObject]]

    @classmethod
    def setdefault(cls, name):
        ''' Set a default implementation. '''
        impl = cls._baseimpl.derived().get(name.upper())
        if impl:
            cls._baseimpl._default = name.upper()
            cls._baseimpl._generator = impl

    @classmethod
    def getdefault(cls):
        ''' Get the default implementation. '''
        default = cls._baseimpl._default
        return cls._baseimpl.derived().get(default) if default else cls._baseimpl

    @classmethod
    def getbase(cls):
        ''' Get the base implementation. '''
        return cls._baseimpl

    @classmethod
    def name(cls):
        ''' Return the name of this implementation. '''
        return cls.__name__.upper()

    @classmethod
    def selectdefault(cls):
        ''' Select the default implementation. '''
        base = cls._baseimpl
        base.derived().get(base._default, base).select()

    @classmethod
    def select(cls, instance=None):
        ''' Select instance class or caller class as generator,
            and update known function object references. '''
        if instance is not None:
            cls._baseimpl._generator = type(instance)
            cls._baseimpl._selinstance = instance
            target = instance
        else:
            cls._baseimpl._generator = cls
            target = cls

        for name, fobj in target.__func_objects__.items():
            if not inspect.ismethod(fobj.func): # TODO: doesn't this disallow e.g., selecting different stack function implementations?
                fobj.update(getattr(target, name, None))

    @classmethod
    def selected(cls):
        ''' Return the selected implementation. '''
        return cls._baseimpl._generator

    @classmethod
    def derived(cls):
        ''' Recursively find all derived classes of cls. '''
        ret = {cls.__name__.upper(): cls}
        for sub in cls.__subclasses__():
            ret.update(sub.derived())
        return ret

    def __init_subclass__(cls, replaceable=True, skipbase=False):
        ''' Register replaceable class bases. '''
        if not skipbase:
            cls._generator = cls
            # Add base functionality
            # Each first implementation descendant of Base
            # is referenced, and keeps a dict of all
            # function-wrapping objects
            if not hasattr(cls, '_baseimpl'):
                cls._baseimpl = cls
                cls._selinstance = None
                cls.__func_objects__ = dict()
    
                # Singleton Entities derive from Replaceable, but have
                # the option to avoid being replaceable.
                if replaceable:
                    cls._default = ''
                    replaceables[cls.__name__.upper()] = cls

            # Always update the function object list by iterating over all items
            for name, func in inspect.getmembers(cls):
                ufunc = inspect.unwrap(func)
                fobj = cls.__func_objects__.get(name)
                if fobj or hasattr(ufunc, '__func_object__'):
                    if fobj is None:
                        fobj = ufunc.__func_object__
                        cls.__func_objects__[name] = fobj
                    else:
                        setattr(ufunc, '__func_object__', fobj)
                    if not inspect.ismethod(fobj.func) and inspect.ismethod(func):
                        # Update callback of function object with first bound method we encounter
                        fobj.update(func)
        return super().__init_subclass__()

    def __init__(self) -> None:
        super().__init__()
        # Make sure that the first constructed instance of the selected class is also selected, 
        # even if select() isn't called explicitly
        cls = type(self)
        if cls._selinstance is None and cls is cls.selected():
            cls.select(self)

    def __new__(cls, *args, **kwargs):
        ''' Replaced new to allow base class to construct selected derived instances. '''
        # Calling the base constructor should return an instance of the
        # selected class. Explicitly calling a subclass constructor should
        # return an instance of that class
        generator = cls._generator if cls is cls.getbase() else cls

        return super().__new__(generator)



""" Console interface for the QTGL implementation."""
from PyQt6.QtCore import Qt, QUrl
from PyQt6.QtGui import QDesktopServices
from PyQt6.QtWidgets import QApplication
from PyQt6.QtWidgets import QWidget, QTextEdit

import bluesky as bs
from bluesky.core.signal import Signal
from bluesky.stack.cmdparser import Command
from bluesky.tools import cachefile
from bluesky.tools.misc import cmdsplit
from bluesky.network import subscribe
from bluesky.network.sharedstate import ActData, get
from bluesky.ui.qtgl import autocomplete
from bluesky.ui.radarclick import radarclick


cmdline_stacked = Signal('cmdline_stacked')


def get_cmd():
    """ Return the current command in the console's command line."""
    if not Console._instance:
        return ''
    return Console._instance.cmd


def get_cmdline():
    """ Return the current text in the console's command line."""
    if not Console._instance:
        return ''
    return Console._instance.command_line


def get_args():
    """ Return the current command arguments in the console's command line."""
    if not Console._instance:
        return []
    return Console._instance.args


def process_cmdline(cmdlines):
    assert Console._instance is not None, 'No console created yet: can only change' + \
        ' command line after main window is created.'
    lines = cmdlines.split('\n')
    if lines:
        Console._instance.append_cmdline(lines[-1])
        for cmd in lines[:-1]:
            Console._instance.stack(cmd)


class Console(QWidget):
    lineEdit: QTextEdit = None
    stackText: QTextEdit = None
    _instance = None

    # Per-remote data
    echotext: ActData[list] = ActData()
    echoflags: ActData[list] = ActData()
    cmddict: ActData[dict] = ActData(group='stackcmds')

    def __init__(self, parent=None):
        super().__init__(parent)
        with cachefile.openfile('console_history.p') as cache:
            try:
                self.command_history = cache.load()
            except:
                self.command_history = []
        self.cmd = ''
        self.args = []
        self.history_pos = 0
        self.command_mem = ''
        self.command_line = ''

        # Connect to the io client's activenode changed signal
        Signal('CMDLINE').connect(self.set_cmdline)
        Signal('actnode-changed').connect(self.actnodeChanged)
        Signal('radarclick').connect(self.on_radarclick)

        assert Console._instance is None, "Console constructor: console instance " + \
            "already exists! Cannot have more than one console."
        Console._instance = self

        # Connect function to save command history on quit
        QApplication.instance().aboutToQuit.connect(self.close)

        # Connect to stack command list SharedState
        subscribe('STACKCMDS')

    def close(self):
        ''' Save command history when BlueSky closes. '''
        with cachefile.openfile('console_history.p') as cache:
            cache.dump(self.command_history)

    def actnodeChanged(self, nodeid):
        text = ('<br>'.join(self.echotext)).replace('\n', '<br>') + '<br>'
        self.stackText.setHtml(text)
        self.stackText.verticalScrollBar().setValue(
            self.stackText.verticalScrollBar().maximum())

    def stack(self, text):
        # Add command to the command history
        self.command_history.append(text)
        self.echo(text)
        bs.stack.stack(text)
        cmdline_stacked.emit(self.cmd, self.args)
        # reset commandline and the autocomplete history
        self.set_cmdline('')
        autocomplete.reset()
        self.history_pos = 0

    def echo(self, text, flags=None):
        cursor = self.stackText.textCursor()
        cursor.clearSelection()
        cursor.movePosition(cursor.MoveOperation.End)
        self.stackText.setTextCursor(cursor)
        self.stackText.insertHtml(text.replace('\n', '<br>') + '<br>')
        self.stackText.verticalScrollBar().setValue(
            self.stackText.verticalScrollBar().maximum())

    def on_radarclick(self, lat, lon):
        actdata = get()
        # TODO routedata isn't really a sharedstate, it only gives a selected route
        tostack, tocmdline = radarclick(get_cmdline(), lat, lon,
                                        actdata.acdata, getattr(actdata, 'routedata', None))

        process_cmdline((tostack + '\n' + tocmdline) if tostack else tocmdline)


    def append_cmdline(self, text):
        self.set_cmdline(self.command_line + text)

    def set_cmdline(self, text, cursorpos=None):
        if self.command_line == text:
            return

        self.command_line = text
        self.cmd, self.args = cmdsplit(self.command_line)
        self.cmd = self.cmd.upper()

        hintline = ''
        guicmd = Command.cmddict.get(self.cmd)
        hint = guicmd.brief[len(self.cmd) + 1:] if guicmd else self.cmddict.get(self.cmd)
        if hint:
            if len(self.args) > 0:
                hintargs = hint.split(',')
                hintline = ' ' + str.join(',', hintargs[len(self.args):])
            else:
                hintline = ' ' + hint

        self.lineEdit.set_cmdline(self.command_line, hintline, cursorpos)

    def keyPressEvent(self, event):
        ''' Handle keyboard input for bluesky. '''
        # Enter-key: enter command
        if event.key() == Qt.Key.Key_Enter or event.key() == Qt.Key.Key_Return:
            if self.command_line:
                # emit a signal with the command for the simulation thread
                self.stack(self.command_line)
                # Clear any shape command preview on the radar display
                # self.radarwidget.previewpoly(None)
                return

        newcmd = self.command_line
        cursorpos = None
        if event.key() >= Qt.Key.Key_Space and event.key() <= Qt.Key.Key_AsciiTilde:
            pos = self.lineEdit.cursor_pos()
            newcmd = newcmd[:pos] + event.text() + newcmd[pos:]
            # Update the cursor position with the length of the added text
            cursorpos = pos + len(event.text())
        elif event.key() == Qt.Key.Key_Backspace:
            pos = self.lineEdit.cursor_pos()
            newcmd = newcmd[:pos - 1] + newcmd[pos:]
            cursorpos = pos - 1
        elif event.key() == Qt.Key.Key_Tab:
            if newcmd:
                newcmd, displaytext = autocomplete.complete(newcmd)
                if displaytext:
                    self.echo(displaytext)
        elif not event.modifiers() & (Qt.KeyboardModifier.ControlModifier | Qt.KeyboardModifier.ShiftModifier | 
                                        Qt.KeyboardModifier.AltModifier | Qt.KeyboardModifier.MetaModifier):
            if event.key() == Qt.Key.Key_Up:
                if self.history_pos == 0:
                    self.command_mem = newcmd
                if len(self.command_history) >= self.history_pos + 1:
                    self.history_pos += 1
                    newcmd = self.command_history[-self.history_pos]

            elif event.key() == Qt.Key.Key_Down:
                if self.history_pos > 0:
                    self.history_pos -= 1
                    if self.history_pos == 0:
                        newcmd = self.command_mem
                    else:
                        newcmd = self.command_history[-self.history_pos]

            elif event.key() == Qt.Key.Key_Left:
                self.lineEdit.cursor_left()

            elif event.key() == Qt.Key.Key_Right:
                self.lineEdit.cursor_right()
            else:
                # Remaining keys are things like sole modifier keys, and function keys
                super().keyPressEvent(event)
        else:
            event.ignore()
            return

        # Final processing of the command line
        self.set_cmdline(newcmd, cursorpos)


class Word:
    def __init__(self, w='', type='str'):
        self.word = w

    def append(self, w):
        self.word = self.word + w
        # TODO do checks

    def __str__(self):
        ''' Word with color markup. '''
        return ''


class Cmdline(QTextEdit):
    ''' Wrapper class for the command line. '''

    def __init__(self, parent=None):
        super().__init__(parent)
        Console.lineEdit = self
        self.cmdline = ''
        # self.setFocusPolicy(Qt.NoFocus)
        self.set_cmdline('')

    def keyPressEvent(self, event):
        event.ignore()

    def set_cmdline(self, cmdline, hints='', cursorpos=None):
        ''' Set the command line with possible hints. '''
        self.setHtml('>>' + cmdline + '<font color="#aaaaaa">' +
                     hints + '</font>')
        self.cmdline = cmdline
        cursor = self.textCursor()
        cursor.setPosition((cursorpos or len(cmdline)) + 2)
        self.setTextCursor(cursor)
        # TODO: word objects with possible list of checker functions?

    def cursor_pos(self):
        ''' Get the cursor position. '''
        return self.textCursor().position() - 2

    def cursor_left(self):
        ''' Move the cursor one position to the left. '''
        cursor = self.textCursor()
        cursor.setPosition(max(2, cursor.position() - 1))
        self.setTextCursor(cursor)

    def cursor_right(self):
        ''' Move the cursor one position to the right. '''
        cursor = self.textCursor()
        cursor.setPosition(min(len(self.cmdline) + 2, cursor.position() + 1))
        self.setTextCursor(cursor)


class Stackwin(QTextEdit):
    ''' Wrapper class for the stack output textbox. '''

    def __init__(self, parent=None):
        super().__init__(parent)
        Console.stackText = self
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)

    def mousePressEvent(self, e):
        self.anchor = self.anchorAt(e.pos())
        if self.anchor:
            QApplication.setOverrideCursor(Qt.CursorShape.PointingHandCursor)

    def mouseReleaseEvent(self, e):
        if self.anchor:
            QDesktopServices.openUrl(QUrl(self.anchor))
            QApplication.setOverrideCursor(Qt.CursorShape.ArrowCursor)
            self.anchor = None



''' Server test '''
import sys
import signal
from subprocess import Popen
from multiprocessing import cpu_count
from threading import Thread

import zmq
import msgpack

import bluesky as bs
from bluesky.network.npcodec import encode_ndarray
from bluesky.network.discovery import Discovery
from bluesky.network.common import genid, bin2hex, MSG_SUBSCRIBE, MSG_UNSUBSCRIBE, GROUPID_SIM, IDLEN


# Register settings defaults
bs.settings.set_variable_defaults(max_nnodes=cpu_count(),
                                  recv_port=11000, send_port=11001,
                                  enable_discovery=False)

def split_scenarios(scentime, scencmd):
    ''' Split the contents of a batch file into individual scenarios. '''
    start = 0
    for i in range(1, len(scencmd) + 1):
        if i == len(scencmd) or scencmd[i][:4] == 'SCEN':
            scenname = scencmd[start].split()[1].strip()
            yield dict(name=scenname, scentime=scentime[start:i], scencmd=scencmd[start:i])
            start = i


class Server(Thread):
    ''' Implementation of the BlueSky simulation server. '''
    def __init__(self, discovery, altconfig=None, startscn=None):
        super().__init__()
        self.spawned_processes = dict()
        self.running = True
        self.max_nnodes = min(cpu_count(), bs.settings.max_nnodes)
        self.scenarios = []
        self.server_id = genid(groupid=GROUPID_SIM, seqidx=0)
        self.max_group_idx = 0
        self.nodes = set()
        self.all_nodes = set()
        self.avail_nodes = set()
        
        # Information to pass on to spawned nodes
        self.altconfig = altconfig
        self.startscn = startscn

        if bs.settings.enable_discovery or discovery:
            self.discovery = Discovery(self.server_id, is_client=False)
        else:
            self.discovery = None

        # Get ZMQ context
        ctx = zmq.Context.instance()
        self.sock_recv = ctx.socket(zmq.XSUB)
        self.sock_send = ctx.socket(zmq.XPUB)
        self.poller = zmq.Poller()

        # Connect to interrupt signal
        signal.signal(signal.SIGINT, lambda *args: self.quit())
        signal.signal(signal.SIGTERM, lambda *args: self.quit())

    def quit(self):
        self.running = False

    def sendscenario(self, worker_id):
        # Send a new scenario to the target sim process
        scen = self.scenarios.pop(0)
        self.send(b'BATCH', scen, worker_id)

    def addnodes(self, count=1, node_ids=None, startscn=None):
        ''' Add [count] nodes to this server. '''
        for idx in range(count):
            if node_ids:
                newid = genid(node_ids[idx])
            else:
                self.max_group_idx += 1
                newid = genid(self.server_id[:-1], seqidx=self.max_group_idx)
            args = [sys.executable, '-m', 'bluesky', '--sim', '--groupid', bin2hex(newid)]
            if self.altconfig:
                args.extend(['--configfile', self.altconfig])
            if startscn:
                args.extend(['--scenfile', startscn])
            self.spawned_processes[newid] = Popen(args)

    def send(self, topic, data='', dest=b''):
        self.sock_send.send_multipart(
            [
                dest.ljust(IDLEN, b'*') + topic + self.server_id,
                msgpack.packb(data, default=encode_ndarray, use_bin_type=True)
            ]
        )

    def run(self):
        ''' The main loop of this server. '''
        print(f'Starting server with id', self.server_id)
        self.sock_recv.bind(f'tcp://*:{bs.settings.send_port}')
        self.sock_send.bind(f'tcp://*:{bs.settings.recv_port}')


        # Create poller for pub/sub and discovery
        self.poller.register(self.sock_recv, zmq.POLLIN)
        self.poller.register(self.sock_send, zmq.POLLIN)

        if self.discovery:
            self.poller.register(self.discovery.handle, zmq.POLLIN)
        print(f'Discovery is {"en" if self.discovery else "dis"}abled')

        # Create subscription for messages targeted at this server
        self.sock_recv.send_multipart([b'\x01' + self.server_id])

        # Start the first simulation node
        self.addnodes(startscn=self.startscn)

        while self.running:
            try:
                events = dict(self.poller.poll(None))
            except zmq.ZMQError:
                print('ERROR while polling')
                break  # interrupted

            # The socket with incoming data
            for sock, event in events.items():
                if event != zmq.POLLIN:
                    # The event does not refer to incoming data: skip for now
                    continue

                # First check if the poller was triggered by the discovery socket
                if self.discovery and sock == self.discovery.handle.fileno():
                    # This is a discovery message
                    dmsg = self.discovery.recv_reqreply()
                    if dmsg.conn_id != self.server_id and dmsg.is_request:
                        # This is a request from someone else: send a reply
                        self.discovery.send_reply(bs.settings.recv_port,
                            bs.settings.send_port)
                    continue

                # Receive the message
                msg = sock.recv_multipart()
                if not msg:
                    # In the rare case that a message is empty, skip remaning processing
                    continue
                if sock == self.sock_send:
                    # This is an (un)subscribe message. If it's an id-only subscription
                    # this is also a registration message
                    if len(msg[0]) == IDLEN + 1:
                        if msg[0][0] == MSG_SUBSCRIBE:
                            # This is an initial client, server, or node subscription
                            if msg[0][1] == GROUPID_SIM and msg[0][1:] in self.spawned_processes:
                                # This is a node owned by this server which has successfully started.
                                self.nodes.add(msg[0][1:])

                        elif msg[0][0] == MSG_UNSUBSCRIBE and msg[0][1] == GROUPID_SIM and msg[0][1:] in self.spawned_processes:
                            print('Removing node', msg[0][1:])
                            self.nodes.discard(msg[0][1:])
                    # Always forward
                    self.sock_recv.send_multipart(msg)
                elif sock == self.sock_recv:
                    # First check if message is directed at this server
                    if msg[0].startswith(self.server_id):
                        topic, sender_id = msg[0][IDLEN:-IDLEN], msg[0][-IDLEN:]
                        data = msgpack.unpackb(msg[1], raw=False)
                        # TODO: also use Signal logic in server?
                        if topic == b'QUIT':
                            self.quit()
                        elif topic == b'ADDNODES':
                            if isinstance(data, int):
                                self.addnodes(count=data)
                            elif isinstance(data, dict):
                                self.addnodes(**data)
                        elif topic == b'STATECHANGE':
                            state = data
                            if state < bs.OP:
                                # If we have batch scenarios waiting, send
                                # the worker a new scenario, otherwise store it in
                                # the available worker list
                                if self.scenarios:
                                    self.sendscenario(sender_id)
                                else:
                                    self.avail_nodes.add(sender_id)
                            else:
                                self.avail_nodes.discard(sender_id)
                        elif topic == b'BATCH':
                            scentime, scencmd = data
                            self.scenarios = [scen for scen in split_scenarios(scentime, scencmd)]
                            # Check if the batch list contains scenarios
                            if not self.scenarios:
                                echomsg = 'No scenarios defined in batch file!'
                            else:
                                echomsg = f'Found {len(self.scenarios)} scenarios in batch'
                                # Send scenario to available nodes (nodes that are in init or hold mode):
                                while self.avail_workers and self.scenarios:
                                    worker_id = next(iter(self.avail_workers))
                                    self.sendscenario(worker_id)
                                    self.avail_workers.pop(worker_id)

                                # If there are still scenarios left, determine and
                                # start the required number of local nodes
                                reqd_nnodes = min(len(self.scenarios), max(0, self.max_nnodes - len(self.workers)))
                                self.addnodes(reqd_nnodes)
                            # ECHO the results to the calling client
                            topic = b'ECHO'
                            data = msgpack.packb(dict(text=echomsg, flags=0), use_bin_type=True)
                            self.sock_send.send_multipart([sender_id + topic + self.server_id, data])
                    else:
                        self.sock_send.send_multipart(msg)
        print('Server quit. Stopping nodes:')
        for pid, p in self.spawned_processes.items():
            print('Stopping node:', pid, end=' ')
            p.send_signal(signal.SIGTERM)
            p.wait()
            # Inform network that node is removed
            self.sock_recv.send_multipart([b'\x00' + pid])
            print('done')
        print('Closing connections:', end=' ')
        self.poller.unregister(self.sock_recv)
        self.poller.unregister(self.sock_send)
        self.sock_recv.close()
        self.sock_send.close()
        zmq.Context.instance().destroy()
        print('done')



'''BADA Coefficient file loader

   This module provides access to the performance data contained in the various
   BADA data files.

   The current implementation is based on the official documentation described in
   report: EEC Technical/Scientific Report No. 14/04/24-44. This report can be obtained here:
   https://www.eurocontrol.int/sites/default/files/field_tabs/content/documents/sesar/user-manual-bada-3-12.pdf
'''
import re
from .fwparser import FixedWidthParser, ParseError
import bluesky as bs

bs.settings.set_variable_defaults(perf_path_bada='performance/BADA')


# File formats of BADA data files. Uses fortran-like notation
# Adapted from the BADA manual format lines. (page 61-81 in the BADA manual)
# Skip characters are indicated with nnX
# Variables are indicated with nnF, nnI, and nnS (float, int, string)
syn_format = ['CD, 1X, 1S, 1X, 4S, 3X, 18S, 1X, 25S, 1X, 6S, 2X, 1S']
syn_parser = FixedWidthParser(syn_format)

opf_format = [
              # aircraft type block (1 data line)
              'CD, 3X, 6S, 9X, 1I, 12X, 9S, 17X, 1S',
              # mass block (1 data line)
              'CD, 2X, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F',
              # flight envelope block (1 data line)
              'CD, 2X, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F',
              # aerodynamics block (12 data lines)
              'CD, 2X, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F',
              'CD, 15X, 3X, 10F, 3X, 10F, 3X, 10F',
              'CD, 15X, 3X, 10F, 3X, 10F, 3X, 10F',
              'CD, 15X, 3X, 10F, 3X, 10F, 3X, 10F',
              'CD, 15X, 3X, 10F, 3X, 10F, 3X, 10F',
              'CD, 15X, 3X, 10F, 3X, 10F, 3X, 10F',
              'CD 50X',
              'CD 50X',
              'CD 50X',
              'CD, 31X, 10F',
              'CD 50X',
              'CD 50X',
              # engine thrust block (3 data lines)
              'CD, 2X, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F',
              'CD, 2X, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F',
              'CD, 2X, 3X, 10F, 3X, 10F',
              # fuel consumption block (3 data lines)
              'CD, 2X, 3X, 10F, 3X, 10F',
              'CD, 2X, 3X, 10F, 3X, 10F',
              'CD, 5X, 10F',
              # ground movement block (1 data line)
              'CD, 2X, 3X, 10F, 3X, 10F, 3X, 10F, 3X, 10F']
opf_parser = FixedWidthParser(opf_format)

apt_format = [
              # company name (1 line)
              'CD, 2X, 3S, 1X, 2S, 4X, 15S',
              # profiles for climb, cruise, and descent (3 lines)
              'CD, 25X, 3I, 1X, 3I, 1X, 2I, 10X, 3I, 1X, 3I, 1X, 2I, 2X, 2I, 1X, 3I, 1X, 3I',
              'CD, 25X, 3I, 1X, 3I, 1X, 2I, 10X, 3I, 1X, 3I, 1X, 2I, 2X, 2I, 1X, 3I, 1X, 3I',
              'CD, 25X, 3I, 1X, 3I, 1X, 2I, 10X, 3I, 1X, 3I, 1X, 2I, 2X, 2I, 1X, 3I, 1X, 3I']
apf_parser = FixedWidthParser(apt_format)

# The available aircraft are stored by type id in synonyms. The actual coefficient data are stored in accoeffs
synonyms     = dict()
accoeffs     = dict()
release_date = 'Unknown'
bada_version = 'Unknown'


def getCoefficients(actype):
    ''' Get a set of BADA coefficients for the given aircraft type.

        This function looks for the given aircraft type in the synonym list, and
        when successful, retreives the corresponding coefficient set.
        This function returns the synonym object (which contains more detailed
        information about the aircraft type) and the coefficient object'''
    syn = synonyms.get(actype, None)
    if syn is None:
        return False, actype + ' is not found in BADA aircraft database. \
            (Check the file SYNONYM.NEW in your BADA path if you spelled the id correctly)'
    coeff = accoeffs.get(syn.file, None)
    if coeff is None:
        return False, actype + ' exists in BADA synonym database, but corresponding \
            coefficient file (%s) could not be found.' % syn.file

    return syn, coeff


def check():
    ''' Import check for BADA performance model. '''
    base = bs.resource(bs.settings.perf_path_bada)
    releasefile = base / 'ReleaseSummary'
    if not releasefile.is_file():
        raise ImportError(f'BADA performance model: Error trying to find BADA files in {base}!')
    synonymfile = base / 'SYNONYM.NEW'
    if not synonymfile.is_file():
        raise ImportError(f'BADA performance model: Error trying to find BADA files in {base}!')
    return True


def init(bada_path=''):
    ''' init() loads the available BADA datafiles in the provided directory.'''
    if accoeffs:
        return True
    bada_path = bs.resource(bs.settings.perf_path_bada)
    releasefile = bada_path / 'ReleaseSummary'
    if releasefile.is_file():
        global release_date, bada_version
        re_reldate = re.compile(r'Summary Date:\s+(.+(?<!\s))\s*', re.IGNORECASE)
        re_badaver = re.compile(r'\s*BADA Release:\s+([\d.]+)\s*', re.IGNORECASE)
        with open(releasefile) as f:
            for line in f:
                if re_reldate.match(line):
                    release_date = re_reldate.findall(line)[0]
                elif re_badaver.match(line):
                    bada_version = re_badaver.findall(line)[0]

                if 'Unknown' not in (release_date, bada_version):
                    break
        print('Found BADA version %s (release date %s)' % (bada_version, release_date))
    else:
        print('No BADA release summary found: can not determine version.')

    synonymfile = bada_path / 'SYNONYM.NEW'
    if not synonymfile.is_file():
        print('SYNONYM.NEW not found in BADA path, could not load BADA.')
        return False

    try:
        data = syn_parser.parse(synonymfile)
    except ParseError as e:
        print(f'Error reading synonym file {e.fname} on line {e.lineno}')
        return False

    for line in data:
        syn = Synonym(line)
        synonyms[syn.accode] = syn
    print('%d aircraft entries loaded' % len(synonyms))

    # Load aircraft coefficient data
    for fname in bada_path.glob('*.OPF'):
        ac = ACData()
        try:
            ac.setOPFData(opf_parser.parse(fname))
            apf = fname.with_suffix('.APF')
            if apf.is_file():
                ac.setAPFData(apf_parser.parse(apf))

        except ParseError as e:
            print(f'Error reading {e.fname} on line {e.lineno}')
            ac = None

        if ac:
            accoeffs[ac.actype] = ac
    print('%d unique aircraft coefficient sets loaded' % len(accoeffs))
    return (len(synonyms) > 0 and len(accoeffs) > 0)


class Synonym:
    def __init__(self, data):
        self.is_equiv = (data[0] == '*')           # False if model is directly supported in bada, true if supported through equivalent model
        self.accode   = data[1]          # Aircraft code
        self.manufact = data[2]         # Aircraft manufacturer
        self.model    = data[3]        # Aircraft model
        self.file     = data[4]        # Corresponding coefficient filename
        self.icao     = (data[5].upper() == 'Y')  # designator for this aircraft type is in use according to ICAO Doc 8643 [RD2]


class ACData:
    # minimum speed coefficients
    CVmin          = 1.3
    CVmin_to       = 1.2

    # reduced power coefficients
    Cred_turboprop = 0.25
    Cred_jet       = 0.15
    Cred_piston    = 0.0

    # value from BADA.gpf file
    gr_acc         = 2.0

    def setOPFData(self, data):
        # aircraft type block:    1 line
        self.actype, self.neng, \
            self.engtype, self.weightcat          = data[0]
        # mass block:             1 line
        self.m_ref, self.m_min, self.m_max, \
            self.m_paymax, self.mass_grad         = data[1]
        # flight envelope block:  1 line
        self.VMO, self.MMO, self.h_MO, \
            self.h_max, self.temp_grad            = data[2]
        # aerodynamics block:     12 lines
        self.S, self.Clbo, self.k, self.CM16      = data[3]
        self.Vstall_cr, self.CD0_cr, self.CD2_cr  = data[4]
        self.Vstall_ic, self.CD0_ic, self.CD2_ic  = data[5]
        self.Vstall_to, self.CD0_to, self.CD2_to  = data[6]
        self.Vstall_ap, self.CD0_ap, self.CD2_ap  = data[7]
        self.Vstall_ld, self.CD0_ld, self.CD2_ld  = data[8]
        self.CD0_gear                             = data[12][0]
        # engine thrust block:    3 lines
        self.CTC                                  = data[15]
        self.CTdes_low, self.CTdes_high, \
            self.Hp_des, self.CTdes_app, \
            self.CTdes_land                       = data[16]
        self.Vdes_ref, self.Mdes_ref              = data[17]
        # fuel consumption block: 3 lines
        self.Cf1, self.Cf2                        = data[18]
        self.Cf3, self.Cf4                        = data[19]
        self.Cf_cruise                            = data[20][0]
        # ground movements block: 1 line
        self.TOL, self.LDL, \
            self.wingspan, self.length            = data[21]

        # Set minimum operating speeds based on stall speeds and minspeed coefficients
        self.vmto = self.Vstall_to * self.CVmin_to
        self.vmic = self.Vstall_ic * self.CVmin
        self.vmcr = self.Vstall_cr * self.CVmin
        self.vmap = self.Vstall_ap * self.CVmin
        self.vmld = self.Vstall_ld * self.CVmin

    def setAPFData(self, data):
        # Minimum, average, and high reference speeds for climb, cruise,
        # and descent. xx1=low mass, xx2=high mass
        self.CAScl1, self.CAScl2, self.Mcl, \
            self.CAScr1, self.CAScr2, self.Mcr, \
            self.Mdes, self.CASdes2, self.CASdes1 = list(zip(*data[1:]))  # swap rows/columns

        # Mach numbers are multiplied by 100 in the BADA files
        self.Mcl  = [m / 100.0 for m in self.Mcl]
        self.Mcr  = [m / 100.0 for m in self.Mcr]
        self.Mdes = [m / 100.0 for m in self.Mdes]



from pathlib import Path
import sys
import pygrib
import datetime
import requests
import numpy as np
import bluesky as bs
from bluesky import stack
from bluesky.core import timed_function
from bluesky.traffic.windsim import WindSim

bs.settings.set_variable_defaults(
    windgfs_url="https://www.ncei.noaa.gov/data/global-forecast-system/access/historical/analysis/")

# nlayer = 23

datadir = Path('')

def init_plugin():
    global datadir
    datadir = bs.resource(bs.settings.data_path) / 'grib'

    if not datadir.is_dir():
        datadir.mkdir()

    global windgfs
    windgfs = WindGFS()

    config = {
        'plugin_name': 'WINDGFS',
        'plugin_type': 'sim'
    }

    return config

class WindGFS(WindSim):
    def __init__(self):
        super().__init__()
        self.year  = 0
        self.month = 0
        self.day   = 0
        self.hour  = 0
        self.lat0  = -90
        self.lon0  = -180
        self.lat1  = 90
        self.lon1  = 180

        # Switch for periodic loading of new GFS data
        self.autoload = True

    def fetch_grb(self, year, month, day, hour, pred=0):
        ym = "%04d%02d" % (year, month)
        ymd = "%04d%02d%02d" % (year, month, day)
        hm = "%02d00" % hour
        pred = "%03d" % pred

        remote_loc = "/%s/%s/gfsanl_3_%s_%s_%s.grb2" % (ym, ymd, ymd, hm, pred)

        fname = "gfsanl_3_%s_%s_%s.grb2" % (ymd, hm, pred)
        fpath = datadir / fname

        remote_url = bs.settings.windgfs_url + remote_loc

        if not fpath.is_file():
            bs.scr.echo("Downloading file, please wait...")
            print("Downloading %s" % remote_url)

            response = requests.get(remote_url, stream=True)

            if response.status_code != 200:
                print("Error. remote data not found")
                return None

            with open(fpath, "wb") as f:
                total_length = response.headers.get('content-length')

                if total_length is None:  # no content length header
                    f.write(response.content)
                else:
                    dl = 0
                    total_length = int(total_length)
                    for data in response.iter_content(chunk_size=4096):
                        dl += len(data)
                        f.write(data)
                        done = int(50 * dl / total_length)
                        sys.stdout.write("\r[%s%s]" % ('=' * done, ' ' * (50-done)) )
                        sys.stdout.flush()

        bs.scr.echo("Download completed.")
        grb = pygrib.open(fpath)

        return grb

    def extract_wind(self, grb, lat0, lon0, lat1, lon1):

        grb_wind_v = grb.select(shortName="v", typeOfLevel=['isobaricInhPa'])
        grb_wind_u = grb.select(shortName="u", typeOfLevel=['isobaricInhPa'])

        lats = np.array([])
        lons = np.array([])
        alts = np.array([])
        vxs = np.array([])
        vys = np.array([])

        for grbu, grbv in zip(grb_wind_u, grb_wind_v):
            level = grbu.level

            if level < 100:  # lesss than 100 hPa, above about 54 k ft
                continue
            else:
                vxs_ = grbu.values
                vys_ = grbv.values

                p = level * 100
                h = (1 - (p / 101325.0)**0.190264) * 44330.76923    # in meters

                lats_ = grbu.latlons()[0].flatten()
                lons_ = grbu.latlons()[1].flatten()
                alts_ = round(h) * np.ones(len(lats_))

                lats = np.append(lats, lats_)
                lons = np.append(lons, lons_)
                alts = np.append(alts, alts_)
                vxs = np.append(vxs, vxs_)
                vys = np.append(vys, vys_)

        lons = (lons + 180) % 360.0 - 180.0     # convert range from 0~360 to -180~180

        lat0_ = min(lat0, lat1)
        lat1_ = max(lat0, lat1)
        lon0_ = min(lon0, lon1)
        lon1_ = max(lon0, lon1)

        mask = (lats >= lat0_) & (lats <= lat1_) & (lons >= lon0_) & (lons <= lon1_)

        data = np.array([lats[mask], lons[mask], alts[mask], vxs[mask], vys[mask]])

        return data

    @stack.command(name='WINDGFS')
    def loadwind(self, lat0: 'lat', lon0: 'lon', lat1: 'lat', lon1: 'lon',
               year: int=None, month: int=None, day: int=None, hour: int=None):
        ''' WINDGFS: Load a windfield directly from NOAA database.

            Arguments:
            - lat0, lon0, lat1, lon1 [deg]: Bounding box in which to generate wind field
            - year, month, day, hour: Date and time of wind data (optional, will use
              current simulation UTC if not specified).
        '''
        self.lat0, self.lon0, self.lat1, self.lon1 =  min(lat0, lat1), \
                              min(lon0, lon1), max(lat0, lat1), max(lon0, lon1)
        self.year = year or bs.sim.utc.year
        self.month = month or bs.sim.utc.month
        self.day = day or bs.sim.utc.day
        self.hour = hour or bs.sim.utc.hour

        # round hour to 3 hours, check if it is a +3h prediction
        self.hour = round(self.hour/3) * 3
        if self.hour in [3, 9, 15, 21]:
            self.hour = self.hour - 3
            pred = 3
        elif self.hour == 24:
            ymd0 = "%04d%02d%02d" % (self.year, self.month, self.day)
            print(ymd0)
            ymd1 = (datetime.datetime.strptime(ymd0, '%Y%m%d') + 
                    datetime.timedelta(days=1))
            self.year  = ymd1.year
            self.month = ymd1.month
            self.day   = ymd1.day    
            self.hour  = 0
            pred = 0
        else:
            pred = 0

        txt = "Loading wind field for %s-%s-%s %s:00..." % (self.year, self.month, self.day, self.hour)
        bs.scr.echo("%s" % txt)

        grb = self.fetch_grb(self.year, self.month, self.day, self.hour, pred)

        if grb is None or self.lat0 == self.lat1 or self.lon0 == self.lon1:
            return False, "Wind data non-existend in area [%d, %d], [%d, %d]. " \
                % (self.lat0, self.lat1, self.lon0, self.lon1) \
                + "time: %04d-%02d-%02d %02d:00" \
                % (self.year, self.month, self.day, self.hour)

        # first clear exisiting wind field
        self.clear()

        # add new wind field
        data = self.extract_wind(grb, self.lat0, self.lon0, self.lat1, self.lon1).T

        data = data[np.lexsort((data[:, 2], data[:, 1], data[:, 0]))] # Sort by lat, lon, alt
        reshapefactor = int((1 + max(self.lat0, self.lat1) - min(self.lat0, self.lat1)) * \
                            (1 + max(self.lon0, self.lon1) - min(self.lon0, self.lon1)))

        lat     = np.reshape(data[:,0], (reshapefactor, -1)).T[0,:]
        lon     = np.reshape(data[:,1], (reshapefactor, -1)).T[0,:]
        veast   = np.reshape(data[:,3], (reshapefactor, -1)).T
        vnorth  = np.reshape(data[:,4], (reshapefactor, -1)).T
        windalt = np.reshape(data[:,2], (reshapefactor, -1)).T[:,0]

        self.addpointvne(lat, lon, vnorth, veast, windalt)        

        return True, "Wind field updated in area [%d, %d], [%d, %d]. " \
            % (self.lat0, self.lat1, self.lon0, self.lon1) \
            + "time: %04d-%02d-%02d %02d:00" \
            % (self.year, self.month, self.day, self.hour)

    @timed_function(name='WINDGFS', dt=3600)
    def update(self):
        if self.autoload:
            _, txt = self.loadwind(self.lat0, self.lon0, self.lat1, self.lon1)
            bs.scr.echo("%s" % txt)


""" OpenAP performance library. """
import json
import pandas as pd
import bluesky as bs


bs.settings.set_variable_defaults(perf_path_openap="performance/OpenAP")

LIFT_FIXWING = 1  # fixwing aircraft
LIFT_ROTOR = 2  # rotor aircraft

ENG_TYPE_TF = 1  # turbofan, fixwing
ENG_TYPE_TP = 2  # turboprop, fixwing
ENG_TYPE_TS = 3  # turboshlft, rotor


class Coefficient:
    def __init__(self):
        # Load synonyms.dat text file into dictionary
        self.synodict = {}
        with open(bs.resource(bs.settings.perf_path_openap) / 'synonym.dat', "r") as f_syno:
            for line in f_syno.readlines():
                if line.count("#") > 0:
                    dataline, comment = line.split("#")
                else:
                    dataline = line.strip("\n")
                acmod, synomod = dataline.split("=")
                acmod = acmod.strip().upper()
                synomod = synomod.strip().upper()

                if acmod == synomod:
                    continue
                self.synodict[acmod] = synomod

        self.acs_fixwing = self._load_all_fixwing_flavor()
        self.engines_fixwing = pd.read_csv(bs.resource(bs.settings.perf_path_openap) / "fixwing/engines.csv", encoding="utf-8")
        self.limits_fixwing = self._load_all_fixwing_envelop()

        self.acs_rotor = self._load_all_rotor_flavor()
        self.limits_rotor = self._load_all_rotor_envelop()

        self.actypes_fixwing = list(self.acs_fixwing.keys())
        self.actypes_rotor = list(self.acs_rotor.keys())

        df = pd.read_csv(bs.resource(bs.settings.perf_path_openap) / "fixwing/dragpolar.csv", index_col="mdl")
        self.dragpolar_fixwing = df.to_dict(orient="index")
        self.dragpolar_fixwing["NA"] = df.mean().to_dict()

    def _load_all_fixwing_flavor(self):
        import warnings

        warnings.simplefilter("ignore")

        # read fixwing aircraft and engine files
        allengines = pd.read_csv(bs.resource(bs.settings.perf_path_openap) / "fixwing/engines.csv", encoding="utf-8")
        allengines["name"] = allengines["name"].str.upper()
        acs = json.load(open(bs.resource(bs.settings.perf_path_openap) / "fixwing/aircraft.json", "r"))
        acs.pop("__comment")
        acs_ = {}

        for mdl, ac in acs.items():
            acengines = ac["engines"]
            acs_[mdl.upper()] = ac.copy()
            acs_[mdl.upper()]["lifttype"] = LIFT_FIXWING
            acs_[mdl.upper()]["engines"] = {}

            for e in acengines:
                e = e.strip().upper()
                selengine = allengines[allengines["name"].str.startswith(e)]
                if selengine.shape[0] >= 1:
                    engine = json.loads(selengine.iloc[-1, :].to_json())
                    acs_[mdl.upper()]["engines"][engine["name"]] = engine

        return acs_

    def _load_all_rotor_flavor(self):
        # read rotor aircraft
        acs = json.load(open(bs.resource(bs.settings.perf_path_openap) / "rotor/aircraft.json", "r"))
        acs.pop("__comment")
        acs_ = {}
        for mdl, ac in acs.items():
            acs_[mdl.upper()] = ac.copy()
            acs_[mdl.upper()]["lifttype"] = LIFT_ROTOR
        return acs_

    def _load_all_fixwing_envelop(self):
        """load aircraft envelop from the model database,
        All unit in SI"""
        limits_fixwing = {}
        for mdl, ac in self.acs_fixwing.items():
            fenv = bs.resource(bs.settings.perf_path_openap) / "fixwing/wrap" / (mdl.lower() + ".txt")

            if fenv.is_file():
                df = pd.read_fwf(fenv).set_index("variable")
                limits_fixwing[mdl] = {}
                limits_fixwing[mdl]["vminto"] = df.loc["to_v_lof"]["min"]
                limits_fixwing[mdl]["vmaxto"] = df.loc["to_v_lof"]["max"]
                limits_fixwing[mdl]["vminic"] = df.loc["ic_va_avg"]["min"]
                limits_fixwing[mdl]["vmaxic"] = df.loc["ic_va_avg"]["max"]
                limits_fixwing[mdl]["vminer"] = min(
                    df.loc["ic_va_avg"]["min"],
                    df.loc["cl_v_cas_const"]["min"],
                    df.loc["cr_v_cas_mean"]["min"],
                    df.loc["de_v_cas_const"]["min"],
                    df.loc["fa_va_avg"]["min"],
                )
                limits_fixwing[mdl]["vmaxer"] = max(
                    df.loc["ic_va_avg"]["max"],
                    df.loc["cl_v_cas_const"]["max"],
                    df.loc["cr_v_cas_mean"]["max"],
                    df.loc["de_v_cas_const"]["max"],
                    df.loc["fa_va_avg"]["max"],
                )
                limits_fixwing[mdl]["vminap"] = df.loc["fa_va_avg"]["min"]
                limits_fixwing[mdl]["vmaxap"] = df.loc["fa_va_avg"]["max"]
                limits_fixwing[mdl]["vminld"] = df.loc["ld_v_app"]["min"]
                limits_fixwing[mdl]["vmaxld"] = df.loc["ld_v_app"]["max"]

                limits_fixwing[mdl]["vmo"] = limits_fixwing[mdl]["vmaxer"]
                limits_fixwing[mdl]["mmo"] = df.loc["cr_v_mach_max"]["opt"]

                limits_fixwing[mdl]["hmax"] = df.loc["cr_h_max"]["opt"] * 1000
                limits_fixwing[mdl]["crosscl"] = df.loc["cl_h_mach_const"]["opt"]
                limits_fixwing[mdl]["crossde"] = df.loc["de_h_cas_const"]["opt"]

                limits_fixwing[mdl]["axmax"] = df.loc["to_acc_tof"]["max"]

                limits_fixwing[mdl]["vsmax"] = max(
                    df.loc["ic_vs_avg"]["max"],
                    df.loc["cl_vs_avg_pre_cas"]["max"],
                    df.loc["cl_vs_avg_cas_const"]["max"],
                    df.loc["cl_vs_avg_mach_const"]["max"],
                )

                limits_fixwing[mdl]["vsmin"] = min(
                    df.loc["ic_vs_avg"]["min"],
                    df.loc["de_vs_avg_after_cas"]["min"],
                    df.loc["de_vs_avg_cas_const"]["min"],
                    df.loc["de_vs_avg_mach_const"]["min"],
                )

        # create envolop based on synonym
        for mdl in self.synodict.keys():
            if mdl not in limits_fixwing:
                limits_fixwing[mdl] = limits_fixwing[self.synodict[mdl]]

        return limits_fixwing

    def _load_all_rotor_envelop(self):
        """load rotor aircraft envelop, all unit in SI"""
        limits_rotor = {}
        for mdl, ac in self.acs_rotor.items():
            limits_rotor[mdl] = {}

            limits_rotor[mdl]["vmin"] = ac["envelop"].get("v_min", -20)
            limits_rotor[mdl]["vmax"] = ac["envelop"].get("v_max", 20)
            limits_rotor[mdl]["vsmin"] = ac["envelop"].get("vs_min", -5)
            limits_rotor[mdl]["vsmax"] = ac["envelop"].get("vs_max", 5)
            limits_rotor[mdl]["hmax"] = ac["envelop"].get("h_max", 2500)

            params = ["v_min", "v_max", "vs_min", "vs_max", "h_max"]
            if set(params) <= set(ac["envelop"].keys()):
                pass
            else:
                warn = f"Warning: Some performance parameters for {mdl} are not found, default values used."
                print(warn)
                bs.scr.echo(warn)

        return limits_rotor



""" Airspace metrics plugin. """
import numpy as np
from bluesky import stack, traf, sim  #, settings, navdb, traf, sim, scr, tools
from bluesky.core import Entity, timed_function
from bluesky.tools import areafilter, datalog, plotter, geo
from bluesky.tools.aero import nm, ft

# Metrics object
metrics = None

class SectorData:
    # Selected traffic data for a/c in a sector
    def __init__(self):
        self.acid = list()
        self.lat0 = np.array([])
        self.lon0 = np.array([])
        self.dist0 = np.array([])

    def id2idx(self, acid):
        # Fast way of finding indices of all ACID's in our list
        tmp = dict((v, i) for i, v in enumerate(self.acid))
        # Return only the indices
        return np.array([tmp.get(acidi, -1) for acidi in acid], dtype=int)

    def get(self, acid):
        # Get lat,lon and distance flown for this a/c
        idx = self.id2idx(acid)
        return self.lat0[idx], self.lon0[idx], self.dist0[idx]

    def delete(self, acid):
        # Remove an aircraft from our sector traffic data list
        idx = np.sort(self.id2idx(acid))
        self.lat0 = np.delete(self.lat0, idx)
        self.lon0 = np.delete(self.lon0, idx)
        self.dist0 = np.delete(self.dist0, idx)
        for i in reversed(idx):
            del self.acid[i]

    def extend(self, acid, lat0, lon0, dist0):
        # Add several a/c to our list:
        # input: lat0,lon0,dist: np arrays, acid: list of strings
        self.lat0 = np.append(self.lat0, lat0)
        self.lon0 = np.append(self.lon0, lon0)
        self.dist0 = np.append(self.dist0, dist0)
        self.acid.extend(acid)

class Metrics(Entity):
    def __init__(self):
        super().__init__()
        # List of sectors known to this plugin.
        self.sectors = list()
        # List of sets of aircraft in each sector
        self.acinside = list()
        # Static Density metric
        self.sectorsd = np.array([], dtype=np.int)
        # Summed pairwise convergence metric
        self.sectorconv = np.array([], dtype=np.float)
        # Route efficiency metric
        self.sectoreff = []

        self.effplot = None

        self.delac = SectorData()

        self.fsd = None
        self.fconv = None
        self.feff = None

    def create(self, n=1):
        pass
        # print(n, 'aircraft created, ntraf =', traf.ntraf)

    def delete(self, idx):
        self.delac.extend(np.array(traf.id)[idx], traf.lat[idx], traf.lon[idx], traf.distflown[idx])
        # n = len(idx) if isinstance(idx, Collection) else 1
        # print(n, 'aircraft deleted, ntraf =', traf.ntraf, 'idx =', idx, 'len(traf.lat) =', len(traf.lat))

    @timed_function(dt=2.5)
    def update(self):
        ''' Periodic update function for metrics calculation. '''
        self.sectorsd = np.zeros(len(self.sectors))
        self.sectorconv = np.zeros(len(self.sectors))
        self.sectoreff = []
        if not traf.ntraf or not self.sectors:
            return

        # Check convergence using CD with large RPZ and tlook
        confpairs, lospairs, inconf, tcpamax, qdr, dist, dcpa, tcpa, tLOS = \
            traf.cd.detect(traf, traf, np.ones(traf.ntraf) * 20 * nm, traf.cd.hpz, np.ones(traf.ntraf) * 3600)

        if confpairs:
            own, intr = zip(*confpairs)
            ownidx = traf.id2idx(own)
            mask = traf.alt[ownidx] > 70 * ft
            ownidx = np.array(ownidx)[mask]
            dcpa = np.array(dcpa)[mask]
            tcpa = np.array(tcpa)[mask]
        else:
            ownidx = np.array([])
    
        sendeff = False
        for idx, (sector, previnside) in enumerate(zip(self.sectors, self.acinside)):
            inside = areafilter.checkInside(sector, traf.lat, traf.lon, traf.alt)

            sectoreff = []
            # Detect aircraft leaving and entering the sector
            previds = set(previnside.acid)
            ids = set(np.array(traf.id)[inside])
            arrived = list(ids - previds)
            left = previds - ids

            # Split aircraft that left the sector in deleted and not deleted
            left_intraf = left.intersection(traf.id)
            left_del = list(left - left_intraf) # Aircraft id's prev inside but deleted
            left_intraf = list(left_intraf) # Aircraft id's that left sector

            # New a/c in sector arr listed by index in arridx
            arridx = traf.id2idx(arrived)
            leftidx = traf.id2idx(left_intraf)

            # Retrieve the current distance flown for arriving and leaving aircraft
            arrdist = traf.distflown[arridx]
            arrlat = traf.lat[arridx]
            arrlon = traf.lon[arridx]

            # Get all a/c ids that left from the set delac
            leftlat, leftlon, leftdist = self.delac.get(left_del)
            leftlat = np.append(leftlat, traf.lat[leftidx])
            leftlon = np.append(leftlon, traf.lon[leftidx])
            leftdist = np.append(leftdist, traf.distflown[leftidx])
            leftlat0, leftlon0, leftdist0 = previnside.get(left_del + left_intraf)
            self.delac.delete(left_del)

            if len(left) > 0:

                # Exclude aircraft where origin = destination for sector efficiency,
                # so require that distance start-end > 10 nm
                q, d = geo.qdrdist(leftlat0, leftlon0, leftlat, leftlon)
                mask = d > 10

                sectoreff = list((leftdist[mask] - leftdist0[mask]) / d[mask] / nm)
                names = np.array(left_del + left_intraf)[mask]

                for name, eff in zip(names, sectoreff):
                    self.feff.write(f'{sim.simt}, {name}, {eff}\n')
                sendeff = True

                # print('{} aircraft left sector {}, distance flown (acid:dist):'.format(len(left), sector))
                # for a, d0, d1, e in zip(left, leftdist0, leftdist, sectoreff):
                #     print('Aircraft {} flew {} meters (eff = {})'.format(a, round(d1-d0), e))

            # Update inside data for this sector
            previnside.delete(left)
            previnside.extend(arrived, arrlat, arrlon, arrdist)

            self.sectoreff.append(sectoreff)

            self.sectorsd[idx] = np.count_nonzero(inside)
            insidx = np.where(np.logical_and(inside, inconf))
            pairsinside = np.isin(ownidx, insidx)
            if len(pairsinside):
                tnorm = np.array(tcpa)[pairsinside] / 300.0
                dcpanorm = np.array(dcpa)[pairsinside] / (5.0 * nm)
                self.sectorconv[idx] = np.sum(
                    np.sqrt(2.0 / tnorm * tnorm + dcpanorm * dcpanorm))
            else:
                self.sectorconv[idx] = 0

            self.fconv.write(f'{sim.simt}, {self.sectorconv[idx]}\n')
            self.fsd.write(f'{sim.simt}, {self.sectorsd[idx]}\n')
        if sendeff:
            self.effplot.send()

    def reset(self):
        if self.fconv:
            self.fconv.close()
        if self.fsd:
            self.fsd.close()
        if self.feff:
            self.feff.close()

    @stack.command(name='METRICS')
    def stackio(self, cmd:'txt', name:'txt'=''):
        ''' Calculate a set of metrics within specified sectors. '''
        print('BLAAA', cmd, name)
        if cmd == 'LIST':
            if not self.sectors:
                return True, 'No registered sectors available'
            return True, 'Registered sectors:', str.join(', ', self.sectors)
        elif cmd == 'ADDSECTOR':
            if name == 'ALL':
                for areaname in areafilter.basic_shapes.keys():
                    self.stackio('ADDSECTOR', areaname)
            # Add new sector to list.
            elif name in self.sectors:
                return False, 'Sector %s already registered.' % name
            elif areafilter.hasArea(name):
                if not self.sectors:
                    self.fconv = open('output/'+stack.get_scenname()+'convergence.csv', 'w')
                    self.fsd = open('output/'+stack.get_scenname()+'density.csv', 'w')
                    self.feff = open('output/'+stack.get_scenname()+'efficiency.csv', 'w')
                    # Create the plot if this is the first sector
                    plotter.plot('metrics.metrics.sectorsd', dt=2.5, title='Static Density',
                                xlabel='Time', ylabel='Aircraft count', fig=1)
                    plotter.plot('metrics.metrics.sectorconv', dt=2.5, title='Summed Pairwise Convergence',
                                xlabel='Time', ylabel='Convergence', fig=2)
                    self.effplot = plotter.Plot('metrics.metrics.sectoreff', title='Route Efficiency', plot_type='boxplot',
                                xlabel='Sector', ylabel='Efficiency', fig=3)
                # Add new area to the sector list, and add an initial inside count of traffic
                self.sectors.append(name)
                self.acinside.append(SectorData())
                plotter.legend(self.sectors, 1)
                return True, 'Added %s to sector list.' % name

            else:
                return False, "No area found with name '%s', create it first with one of the shape commands" % name

        else:
            # Remove area from sector list
            if name in self.sectors:
                idx = self.sectors.index(name)
                self.sectors.pop(idx)
                return True, 'Removed %s from sector list.' % name
            return False, "No sector registered with name '%s'." % name

def init_plugin():
    # Addtional initilisation code
    global metrics
    metrics = Metrics()
    # Configuration parameters
    config = {
        'plugin_name': 'METRICS',
        'plugin_type': 'sim'
        }

    return config



""" Keyboard and mouse processing for the pygame version of BlueSky."""
from math import *
import pygame as pg
import bluesky as bs
from bluesky import stack
from ..radarclick import radarclick


class Keyboard:
    """
    Keyboard class definition : keyboard & mouse input processing class

    Methods:
        Keyboard(tmx)                      :  constructor

        update()                           : Check for input & process it

    Created by  : Jacco M. Hoekstra (TU Delft)
    """
    def __init__(self):

        self.dragedit = False  # Edit window initially not connected to mouse
        self.dragpotmenu = False
        self.dragmenu = False
        self.dragdx = 0
        self.dragdy = 0
        self.lastcmd  = ""     # previously typed command line
        self.firstx = True
        return

    def update(self):

        # First time: IC window in pygame version
        if self.firstx:
            stack.stack("IC")
            self.firstx = False

        # Get events
        for event in pg.event.get():
            if event.type==pg.QUIT:
                bs.sim.quit()

            elif event.type==pg.KEYDOWN:

                # Alphanumeric key
                if event.key>31 and event.key<127:
                    bs.scr.editwin.insert(str(event.unicode))  #.upper())

                elif event.key==13: # ENTER
                    cmdline = bs.scr.editwin.getline()
                    bs.scr.editwin.enter()
                    if len(cmdline)>0:
                        self.lastcmd = cmdline
                        stack.stack(cmdline)

                elif event.key==8:   # BACKSPACE
                    bs.scr.editwin.backspace()

                elif event.key==27: # ESCAPE
                    bs.sim.quit()

                elif event.key ==pg.K_F3: # F3: Insert last command
                    bs.scr.editwin.insert(self.lastcmd.strip()+" ") # strip().upper()+" ")

                # Display keys
                elif event.key == 269: # Num lock minus
                    stack.stack("ZOOM OUT")
                elif event.key == 270: # Num lock pluus
                    stack.stack("ZOOM IN")

                elif event.key == 273: # Cursor up
                    bs.scr.pan("ABOVE")
                elif event.key == 274: # Cursor down
                    bs.scr.pan("DOWN")
                elif event.key == 275: # Num lock down
                    bs.scr.pan("RIGHT")
                elif event.key == 276: # Num lock right
                    bs.scr.pan("LEFT")

                elif event.key ==pg.K_F11: # F11: Toggle full screen

                    bs.scr.swfullscreen = not bs.scr.swfullscreen
                    bs.scr.fullscreen(bs.scr.swfullscreen)

                else: #TEST OPTION
                    pass

                # bs.scr.editwin.insert(str(event.key))
                # TBD
                # bs.scr.insedit(chr(ikey))

            # End of keys selection


            # Mouse events:
            #    MOUSEMOTION      pos, rel, buttons
            #    MOUSEBUTTONUP    pos, button
            #    MOUSEBUTTONDOWN  pos, button

            # Mouse button 1 release: enter value in edit line
            elif event.type==pg.MOUSEBUTTONUP:

                # Reselase edit window if necessary
                if event.button==1:
                    if self.dragedit:
                        bs.scr.editwin.winx = event.pos[0]-self.dragdx
                        bs.scr.editwin.winy = event.pos[1]-self.dragdy
                        bs.scr.editwin.winy = bs.scr.editwin.fontedit.linedy *       \
                              (bs.scr.editwin.winy/bs.scr.editwin.fontedit.linedy)
                        bs.scr.editwin.rect = pg.Rect(bs.scr.editwin.winx,bs.scr.editwin.winy, \
                             bs.scr.editwin.bmpdx,bs.scr.editwin.bmpdy)
                        bs.scr.redrawedit = True

                    # Menu button click
                    elif bs.scr.menu.rect.collidepoint(event.pos) and \
                         not self.dragmenu:
                        cmdtxt = bs.scr.menu.getcmd(event.pos)
                        if cmdtxt != "":
                            stack.stack(cmdtxt)

                    # In all other cases process as radar click
                    elif self.dragmenu:
                        self.dragmenu    = False
                        self.dragpotmenu = False

                    else:
                        if not bs.scr.swnavdisp:
                            # Interpret current edit line
                            cmdline = bs.scr.editwin.getline()
                            lat, lon = bs.scr.xy2ll(event.pos[0], event.pos[1])
                            tostack, todisplay = radarclick(cmdline, lat, lon)
                            if len(todisplay) > 0:
                                if todisplay[0] == '\n':
                                    bs.scr.editwin.enter()
                                bs.scr.editwin.insert(todisplay.strip('\n'))
                                if todisplay[-1] == '\n':
                                    bs.scr.editwin.enter()
                            if len(tostack) > 0:
                                stack.stack(tostack)

                # Make sure edit and menu window are released
                self.dragedit    = False
                self.dragmenu    = False
                self.dragpotmenu = False

            # Mouse button down: lock onto edit window if insied edit window
            elif event.type==pg.MOUSEBUTTONDOWN:

                self.dragmenu = False
                self.dragpotmenu = False

                if event.button==1:
                    if bs.scr.editwin.rect.collidepoint(event.pos):
                        self.dragedit = True
                        self.dragdx = event.pos[0]-bs.scr.editwin.winx
                        self.dragdy = event.pos[1]-bs.scr.editwin.winy
                        bs.scr.redrawedit = True

                    elif bs.scr.menu.rect.collidepoint(event.pos):
                        self.dragpotmenu = True
                        self.dragmenu    = False
                        bs.scr.redrawedit   = True
                        self.dragdx = event.pos[0] - bs.scr.menu.x
                        self.dragdy = event.pos[1] - bs.scr.menu.y
                elif event.button==4:
                    bs.scr.zoom(sqrt(2.0))
                elif event.button==5:
                    bs.scr.zoom(0.5*sqrt(2.0))

            # Mouse motion: drag edit/menu window with mouse, if necessary
            # Check also for mouse button 1
            elif event.type==pg.MOUSEMOTION and \
                (self.dragedit or self.dragpotmenu or self.dragmenu):
                if self.dragedit:
                    pressed = pg.mouse.get_pressed()[0]
                    if not pressed:
                        self.dragedit = False
                    else:
                        bs.scr.editwin.winx = event.pos[0]-self.dragdx
                        bs.scr.editwin.winy = event.pos[1]-self.dragdy
                        bs.scr.editwin.rect = pg.Rect(bs.scr.editwin.winx,bs.scr.editwin.winy, \
                                 bs.scr.editwin.bmpdx,bs.scr.editwin.bmpdy)
                        bs.scr.redrawedit = True

                elif self.dragpotmenu:
                    pressed = pg.mouse.get_pressed()[0]
                    if  not pressed:
                        self.dragpotmenu = False
                        self.dragmenu = False
                    else:
                        mx,my = pg.mouse.get_pos()
                        outside = not bs.scr.menu.rect.collidepoint((mx,my))
                        if outside and pressed:
                            self.dragmenu = True
                            self.dragpotmenu = False

                elif self.dragmenu:

                    mx,my = event.pos
                    bs.scr.menu.x = mx - self.dragdx
                    bs.scr.menu.y = my - self.dragdy
                    bs.scr.menu.update() # Update rectangle
                    bs.scr.redrawedit = True
                    pressed = pg.mouse.get_pressed()[0]
                    if  not pressed:
                        self.dragpotmenu = False
                        self.dragmenu = False

        #----- End of Update -----
        return



from pathlib import Path
import subprocess

import bluesky as bs
from bluesky.stack.stackbase import Stack, forward, stack
from bluesky.stack.cmdparser import Command, command, commandgroup
from bluesky.stack import argparser

def init():
    ''' client-side stack initialisation. '''
    # Display Help text on start of program
    stack("ECHO BlueSky Console Window: Enter HELP or ? for info.\n"
          "Or select IC to Open a scenario file.")


def process():
    ''' Client-side stack processing. '''
    # Process stack of commands
    for cmdline in Stack.commands():
        success = True
        echotext = ''
        echoflags = bs.BS_OK

        # Get first argument from command line and check if it's a command
        cmd, argstring = argparser.getnextarg(cmdline)
        cmdu = cmd.upper()
        cmdobj = Command.cmddict.get(cmdu)

        # Proceed if a command object was found
        if cmdobj:
            try:
                # Call the command, passing the argument string
                success, echotext = cmdobj(argstring)
                if not success:
                    if not argstring:
                        echotext = echotext or cmdobj.brieftext()
                    else:
                        echoflags = bs.BS_FUNERR
                        echotext = f'Syntax error: {echotext or cmdobj.brieftext()}'

            except Exception as e:
                success = False
                echoflags = bs.BS_ARGERR
                header = '' if not argstring else e.args[0] if e.args else 'Argument error.'
                echotext = f'{header}\nUsage:\n{cmdobj.brieftext()}'

        elif Stack.sender_id is None:
            # If sender_id is None, this stack command originated from the gui. Send it on to the sim
            forward()
        # -------------------------------------------------------------------
        # Command not found
        # -------------------------------------------------------------------
        else:
            success = False
            echoflags = bs.BS_CMDERR
            if not argstring:
                echotext = f'Unknown command or aircraft: {cmd}'
            else:
                echotext = f'Unknown command: {cmd}'

        # Always return on command
        if echotext:
            bs.scr.echo(echotext, echoflags, Stack.sender_id)

    # Clear the processed commands
    Stack.cmdstack.clear()


@commandgroup(name='HELP', aliases=('?',))
def showhelp(cmd: 'txt' = '', subcmd: 'txt' = ''):
    """ HELP: Display general help text or help text for a specific command,
        or dump command reference in file when command is >filename.

        Arguments:
        - cmd: Argument can refer to:
            - Command name to display help for. 
            - Call HELP >filename to generate a CSV file with help text for all commands.
            - Call HELP PDF to view a pdf containing information on all stack commands.

        To get more detailed information on a command type DOC cmd.
    """

    # Check if help is asked for a specific command
    cmdobj = Command.cmddict.get(cmd or 'HELP')
    if cmdobj:
        return True, cmdobj.helptext(subcmd)

    # If command is not a known Client command pass the help request on to the sim
    forward(target_id=bs.net.act_id)


@showhelp.subcommand
def pdf():
    ''' Open a pdf file with BlueSky command help text. '''
    pdfhelp = Path("BLUESKY-COMMAND-TABLE.pdf")
    if (Path('docs') / pdfhelp).is_file():
        try:
            subprocess.Popen(pdfhelp, shell=True, cwd='docs')
        except:
            return "Opening " + pdfhelp.as_posix() + " failed."
    else:
        return pdfhelp.as_posix() + "does not exist."

    return "Pdf window opened"


@command
def makedoc():
    ''' MAKEDOC: Make markdown templates for all stack functions
        that don't have a doc page yet.
    '''
    tmp = Path('tmp')
    if not tmp.is_dir():
        tmp.mkdir()
    # Get unique set of commands
    cmdobjs = set(Command.cmddict.values())
    for o in cmdobjs:
        if not bs.resource(f"html/{o.name}.html").is_file():
            with open(tmp / f"{o.name.lower()}.md", "w") as f:
                f.write(
                    f"# {o.name}: {o.name.capitalize()}\n"
                    + o.help
                    + "\n\n"
                    + "**Usage:**\n\n"
                    + f"    {o.brief}\n\n"
                    + "**Arguments:**\n\n"
                )
                if not o.params:
                    f.write("This command has no arguments.\n\n")
                else:
                    f.write(
                        "|Name|Type|Optional|Description\n"
                        + "|--------|------|---|---------------------------------------------------\n"
                    )
                    for arg in o.params:
                        f.write(str(arg).replace(':', '|') + f" |{arg.hasdefault()}|\n")
                f.write("\n[[Back to command reference.|Command Reference]]\n")
    # To also get all of the sim stack commands, forward to sim
    forward()


import pygame as pg

import bluesky as bs

def show():
    # Show splash screen
    pg.init()
    imgpath = bs.resource("graphics/splash.gif")
    splashimg = pg.image.load(imgpath)
    splashwin = pg.display.set_mode(splashimg.get_size(),pg.NOFRAME)
    splashwin.blit(splashimg,(0,0))
    pg.display.flip()
    return

def destroy():
    pg.display.quit()               # Kill splash screen
    return



import inspect
import functools
from types import SimpleNamespace
from collections import OrderedDict

from bluesky.core.funcobject import FuncObject
from bluesky.core.simtime import Timer


class _Hook(OrderedDict):
    def trigger(self):
        for callback in self.values():
            callback()


# Dictionaries of timed functions for different trigger points
hooks = SimpleNamespace(
    update=_Hook(),
    preupdate=_Hook(),
    hold=_Hook(),
    reset=_Hook()
)


def timed_function(func=None, name='', dt=0, hook='update'):
    ''' Decorator to turn a function into a (periodically) timed function. '''
    def deco(func):
        # Generate a name if none is provided
        if not name:
            if inspect.ismethod(func):
                if inspect.isclass(func.__self__):
                    # classmethods
                    tname = f'{func.__self__.__name__}.{func.__name__}'
                else:
                    # instance methods
                    tname = f'{func.__self__.__class__.__name__}.{func.__name__}'
            else:
                tname = f'{func.__module__}.{func.__name__}'
        else:
            tname = name
        fobj = FuncObject(func)
        timer = None if hook == 'reset' else Timer(tname, dt)

        # Generate the wrapped timed callback
        # Create the appropriate call method
        if 'dt' in inspect.signature(func).parameters:
            # Callback for functions that have dt as argument
            @functools.wraps(fobj)
            def callback(*args):
                if timer.counter == 0:
                    fobj(*args, dt=float(timer.dt_act))
        else:
            # Callback for functions without dt as argument
            @functools.wraps(fobj)
            def callback(*args):
                if timer.counter == 0:
                    fobj(*args)

        # Add automatically-triggered function to appropriate dict(s) if not there yet.
        hooknames = hook if isinstance(hook, (list, tuple)) else (hook,)
        for hookname in hooknames:
            target = getattr(hooks, hookname)
            if target is None:
                raise KeyError(f'No timing hook found with name {hookname}')
            if tname not in target:
                target[tname] = fobj if hookname in ('reset', 'hold') else callback

        # Construct the timed function, but return the original function
        return func

    # Allow both @timed_function and @timed_function(args)
    return deco(func) if func else deco



from bluesky.stack.importer import Importer


### Initialization function of your plugin. Do not change the name of this
### function, as it is the way BlueSky recognises this file as a plugin.
def init_plugin():
    ''' Plugin initialisation function. '''
    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'IMPORTEX',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim',
        }

    # init_plugin() should always return a configuration dict.
    return config


class ImportExample(Importer):
    def __init__(self):
        super().__init__(filetype='Example', extensions=('txt', 'dat', 'log'))

    def load(self, fname):
        # Our file loader should return two lists:
        # - a list of timestamps (in (fractions of) seconds). This list is optional,
        #   for untimestamped commands use an empty list
        # - a list of stack commands
        return [], ['MCRE 1']



''' Entity is a base class for all BlueSky singleton entities. '''
import inspect
from typing import Optional
from bluesky.core.base import Base
from bluesky.core.trafficarrays import TrafficArrays


class Proxy:
    ''' Proxy class for BlueSky replaceable singleton entities. '''
    def __init__(self):
        # The reference object
        self.__dict__['_refobj'] = None
        # List of all proxied functions/methods
        self.__dict__['_proxied'] = list()

    def _selected(self):
        return self._refobj.__class__

    def _replace(self, refobj):
        # Replace our reference object
        self.__dict__['_refobj'] = refobj
        # First clear all proxied functions/methods
        for name in self._proxied:
            delattr(self, name)
        self._proxied.clear()
        # Copy all public functions/methods of reference object
        for name, value in inspect.getmembers(refobj, callable):
            if name[0] != '_':
                self.__dict__[name] = value
                self._proxied.append(name)

    def __getattr__(self, attr):
        return getattr(self._refobj, attr)

    def __setattr__(self, name, value):
        return setattr(self._refobj, name, value)


def isproxied(obj):
    ''' Returns true if 'obj' is a proxied object. '''
    return isinstance(obj, Proxy)


def getproxied(obj):
    ''' Return wrapped proxy object if proxied, otherwise the original object. '''
    return obj.__dict__['_refobj'] if isinstance(obj, Proxy) else obj


class EntityMeta(type):
    ''' Meta class to make replaceable classes singletons. '''
    def __call__(cls, *args, **kwargs):
        ''' Object creation with proxy wrapping and ensurance of singleton
            behaviour. '''
        # Create singleton instance of this class if it doesn't exist yet
        if not cls.is_instantiated():
            super().__call__(*args, **kwargs)

        # When proxied, calling the base constructor returns the proxy object.
        # All derived constructors, and all non-proxied classes return the
        # actual instance.
        if cls._proxy and cls is cls.getbase():
            if getproxied(cls._proxy) is None:
                cls.select(cls._instance)
            return cls._proxy

        return cls._instance


class Entity(Base, TrafficArrays, metaclass=EntityMeta, skipbase=True):
    ''' Super class for BlueSky singleton entities (such as Traffic, Autopilot, ...). '''
    # Class variables that will be set for subclasses
    _proxy: Optional[Proxy]
    _instance: Optional['Entity']

    @classmethod
    def select(cls, instance=None):
        ''' Select this class as generator, and update
            function object references to those of the selected instance. '''
        instance = instance or cls._instance or cls()
        super().select(instance)
        if cls._proxy and cls._proxy._selected() is not cls:
            cls._proxy._replace(instance)


    @classmethod
    def is_instantiated(cls):
        ''' Returns true if the singleton of this class has already been instantiated. '''
        return cls._instance is not None

    @classmethod
    def instance(cls):
        ''' Return the current instance of this entity. '''
        return cls._proxy or cls._instance

    @classmethod
    def implinstance(cls):
        ''' Return the instance of this specific implementation. '''
        return cls._instance

    def __init__(self) -> None:
        super().__init__()
        cls = type(self)
        if cls._instance is None:
            cls._instance = self

    def __init_subclass__(cls, replaceable=False, skipbase=False):
        super().__init_subclass__(replaceable, skipbase)
        # When skipbase is True, an intermediate base class is currently defined,
        # and instance management shoud skip one step in the class tree.
        if skipbase:
            return

        # Each Entity subclass keeps its own (single) instance.
        cls._instance = None
        if not hasattr(cls, '_proxy'):
            # Each first descendant of replaceable Entities has a proxy object
            # that wraps the currently selected instance
            cls._proxy = Proxy() if replaceable else None



# -*- coding: utf-8 -*-

import bluesky as bs
from .misc import txt2lat, txt2lon

def txt2pos(name, reflat, reflon):
    pos = Position(name.upper().strip(), reflat, reflon)
    if not pos.error:
        return True, pos
    return False, name+" not found in database"

def islat(txt):
    # Is it a latitude-like format or not?

    # Take out non-digit chars which are allowed
    testtxt = txt.upper().strip().strip("-")   \
                         .strip("+").strip("\n").strip(",") \
                         .replace('"',"").replace("'","").replace(".","")

    # Take away one leading N or S if present before other chars
    if (testtxt[0]=="N" or testtxt[0]=="S") and len(testtxt)>1:
        testtxt =testtxt[1:]

    try:
        float(testtxt)
    except ValueError:
        return False
    return True

class Position():
    """ Position class: container for position data
    """

    # position types: "latlon","nav","apt","rwy"

    # Initialize using text
    def __init__(self, name, reflat, reflon):

        self.name = name # default: copy source name
        self.error = False # we're optmistic about our succes
        self.refhdg = None

        # lat,lon type ?
        if name.count(",") > 0: #lat,lon or apt,rwy type
            txt1, txt2 = name.split(",")
            if islat(txt1):
                self.lat = txt2lat(txt1)
                self.lon = txt2lon(txt2)
                self.name = ""
                self.type ="latlon"

        # runway type ? "EHAM/RW06","EHGG/RWY27"
        elif name.count("/RW") > 0:
            try:
                aptname, rwytxt = name.split("/RW")
                rwyname = rwytxt.lstrip("Y").upper() # remove Y and spaces
                self.lat, self.lon, self.refhdg = bs.navdb.rwythresholds[aptname][rwyname]
            except KeyError:
                self.error = True
            self.type = "rwy"

        # airport?
        elif bs.navdb.aptid.count(name)>0:
            idx = bs.navdb.aptid.index(name.upper())

            self.lat = bs.navdb.aptlat[idx]
            self.lon = bs.navdb.aptlon[idx]
            self.type ="apt"

        # fix or navaid?
        elif bs.navdb.wpid.count(name) > 0:
            idx = bs.navdb.getwpidx(name, reflat, reflon)
            self.lat = bs.navdb.wplat[idx]
            self.lon = bs.navdb.wplon[idx]
            self.type = "nav"

        # aircraft id?
        elif name in bs.traf.id:
            idx = bs.traf.id2idx(name)
            self.name = ""
            self.type = "latlon"
            self.lat = bs.traf.lat[idx]
            self.lon = bs.traf.lon[idx]

            # exception for pan, check for LEFT, RIGHT, ABOVE or DOWN
        elif name.upper() in ["LEFT","RIGHT","ABOVE","DOWN"]:
            self.lat = reflat
            self.lon = reflon
            self.type = "dir"

# Not used now, but save this code for future use
#            # Make a N52E004 type waypoint name
#            clat = "SN"[lat>0]
#            clon = "WE"[lon>0]
#            name = clat + "%02d"%int(abs(round(lat))) + \
#                   clon + "%03d"%int(abs(round(lon)))
        else:
            self.error = True
            # raise error with missing data... (empty position object)



""" BlueSky ADS-B datafeed plugin. Reads the feed from a Mode-S Beast server,
    and visualizes traffic in BlueSky."""
import time
from bluesky import stack, settings, traf
from bluesky.tools.network import TcpSocket
from bluesky.tools import aero
import bluesky.plugins.adsb_decoder as decoder

## Default settings
# Mode-S / ADS-B server hostname/ip, and server port
settings.set_variable_defaults(modeS_host='', modeS_port=0)

# Global data
reader = None

### Initialization function of the adsbfeed plugin.
def init_plugin():
    # Initialize Modesbeast reader
    global reader
    reader = Modesbeast()

    # Configuration parameters
    config = {
        'plugin_name':     'DATAFEED',
        'plugin_type':     'sim',
        'update_interval': 0.0,
        'preupdate':       reader.update
        }

    stackfunctions = {
        "DATAFEED": [
            "DATAFEED [ON/OFF]",
            "[onoff]",
            reader.toggle,
            "Select an ADS-B data source for traffic"]}

    # init_plugin() should always return these two dicts.
    return config, stackfunctions


class Modesbeast(TcpSocket):
    def __init__(self):
        super().__init__()
        self.acpool         = {}
        self.buffer         = b''
        self.default_ac_mdl = "B738"

    def processData(self, data):
        self.buffer += data

        if len(self.buffer) > 2048:
            # process the buffer until the last divider <esc> 0x1a
            # then, reset the buffer with the remainder

            bfdata = self.buffer
            n = (len(bfdata) - 1) - bfdata[::-1].index(0x1a)
            data = bfdata[:n - 1]
            self.buffer = self.buffer[n:]

            messages = self.read_mode_s(data)

            if not messages:
                return

            for msg, ts in messages:
                self.read_message(msg, ts)
        return

    def read_mode_s(self, data):
        '''
        <esc> "1" : 6 byte MLAT timestamp, 1 byte signal level,
            2 byte Mode-AC
        <esc> "2" : 6 byte MLAT timestamp, 1 byte signal level,
            7 byte Mode-S short frame
        <esc> "3" : 6 byte MLAT timestamp, 1 byte signal level,
            14 byte Mode-S long frame
        <esc> "4" : 6 byte MLAT timestamp, status data, DIP switch
            configuration settings (not on Mode-S Beast classic)
        <esc><esc>: true 0x1a
        <esc> is 0x1a, and "1", "2" and "3" are 0x31, 0x32 and 0x33

        timestamp:
        wiki.modesbeast.com/Radarcape:Firmware_Versions#The_GPS_timestamp
        '''

        # split raw data into chunks
        chunks    = []
        separator = 0x1a
        piece     = []
        for d in data:
            if d == separator:
                # shortest msgs are 11 chars
                if len(piece) > 10:
                    chunks.append(piece)
                piece = []
            piece.append(d)

        # extract messages
        messages = []
        for cnk in chunks:
            msgtype = cnk[1]

            # Mode-S Short Message, 7 byte
            if msgtype == 0x32:
                msg = ''.join('%02X' % i for i in cnk[9:16])

            # Mode-S Short Message, 14 byte
            elif msgtype == 0x33:
                msg = ''.join('%02X' % i for i in cnk[9:23])

            # Other message tupe
            else:
                continue

            ts = time.time()

            messages.append([msg, ts])
        return messages

    def read_message(self, msg, ts):
        """
        Process ADSB messages
        """

        if len(msg) < 28:
            return

        df = decoder.get_df(msg)

        if df == 17:
            addr = decoder.get_icao_addr(msg)
            tc = decoder.get_tc(msg)

            if tc >= 1 and tc <= 4:
                # aircraft identification
                callsign = decoder.get_callsign(msg)
                self.update_callsign(addr, callsign)
            if tc >= 9 and tc <= 18:
                # airbone postion frame
                alt = decoder.get_alt(msg)
                oe = decoder.get_oe_flag(msg)  # odd or even frame
                cprlat = decoder.get_cprlat(msg)
                cprlon = decoder.get_cprlon(msg)
                self.update_cprpos(addr, oe, ts, alt, cprlat, cprlon)
            elif tc == 19:        # airbone velocity frame
                sh = decoder.get_speed_heading(msg)
                if len(sh) == 2:
                    spd = sh[0]
                    hdg = sh[1]
                    self.update_spd_hdg(addr, spd, hdg)
        return

    def update_cprpos(self, addr, oe, ts, alt, cprlat, cprlon):
        if addr in self.acpool:
            ac = self.acpool[addr]
        else:
            ac = {}

        ac['alt'] = alt
        if oe == '1':       # odd frame cpr position
            ac['cprlat1'] = cprlat
            ac['cprlon1'] = cprlon
            ac['t1'] = ts

        if oe == '0':       # even frame cpr position
            ac['cprlat0'] = cprlat
            ac['cprlon0'] = cprlon
            ac['t0'] = ts

        ac['ts'] = time.time()

        self.acpool[addr] = ac
        return

    def update_spd_hdg(self, addr, spd, hdg):
        if addr in self.acpool:
            ac = self.acpool[addr]
        else:
            ac = {}

        ac['speed'] = spd
        ac['heading'] = hdg
        ac['ts'] = time.time()

        self.acpool[addr] = ac
        return

    def update_callsign(self, addr, callsign):
        if addr not in self.acpool:
            self.acpool[addr] = {}

        self.acpool[addr]['callsign'] = callsign
        return

    def update_all_ac_postition(self):
        keys = ('cprlat0', 'cprlat1', 'cprlon0', 'cprlon1')
        for addr, ac in list(self.acpool.items()):
            # check if all needed keys are in dict
            if set(keys).issubset(ac):
                pos = decoder.cpr2position(
                    ac['cprlat0'], ac['cprlat1'],
                    ac['cprlon0'], ac['cprlon1'],
                    ac['t0'], ac['t1']
                )

                # update positions of all aircrafts in the list
                if pos:
                    self.acpool[addr]['lat'] = pos[0]
                    self.acpool[addr]['lon'] = pos[1]
        return

    def stack_all_commands(self):
        """create and stack command"""
        params = ('lat', 'lon', 'alt', 'speed', 'heading', 'callsign')
        for i, d in list(self.acpool.items()):
            # check if all needed keys are in dict
            if set(params).issubset(d):
                acid = d['callsign']
                # check is aircraft is already beening displayed
                if(traf.id2idx(acid) < 0):
                    mdl = self.default_ac_mdl
                    v = aero.tas2cas(d['speed'], d['alt'] * aero.ft)
                    cmdstr = 'CRE %s, %s, %f, %f, %f, %d, %d' % \
                        (acid, mdl, d['lat'], d['lon'],
                            d['heading'], d['alt'], v)
                    stack.stack(cmdstr)
                else:
                    cmdstr = 'MOVE %s, %f, %f, %d' % \
                        (acid, d['lat'], d['lon'], d['alt'])
                    stack.stack(cmdstr)

                    cmdstr = 'HDG %s, %f' % (acid,  d['heading'])
                    stack.stack(cmdstr)

                    v_cas = aero.tas2cas(d['speed'], d['alt'] * aero.ft)
                    cmdstr = 'SPD %s, %f' % (acid,  v_cas)
                    stack.stack(cmdstr)
        return

    def remove_outdated_ac(self):
        """House keeping, remove old entries (offline > 100s)"""
        for addr, ac in list(self.acpool.items()):
            if 'ts' in ac:
                # threshold, remove ac after 90 seconds of no-seen
                if (int(time.time()) - ac['ts']) > 100:
                    del self.acpool[addr]
                    # remove from sim traffic
                    if 'callsign' in ac:
                        stack.stack('DEL %s' % ac['callsign'])
        return

    def debug(self):
        addlist = str.join(', ', self.acpool.keys())
        print(addlist)
        print("")
        print("total count: %d" % len(self.acpool.keys()))
        return

    def update(self):
        if self.isConnected():
            # self.debug()
            self.remove_outdated_ac()
            self.update_all_ac_postition()
            self.stack_all_commands()

    def toggle(self, flag=None):
        if flag is None:
            if self.isConnected():
                return True, 'Connected to %s on port %s' % (settings.modeS_host, settings.modeS_port)
            else:
                return True, 'Not connected'
        elif flag:
            self.connectToHost(settings.modeS_host, settings.modeS_port)
            stack.stack('OP')
            return True, 'Connecting to %s on port %s' % (settings.modeS_host, settings.modeS_port)
        else:
            self.disconnectFromHost()

        return True



from math import cos, radians
from PyQt6.QtWidgets import QListView, QStyledItemDelegate, QStyle, QApplication as app
from PyQt6.QtGui import QColor, QPalette
from PyQt6.QtCore import QSize, Qt, QRect, QAbstractListModel, QModelIndex, Qt, QVariant

from bluesky.network import context as ctx
from bluesky.network.subscriber import subscriber
from bluesky.network.sharedstate import ActData
from bluesky.core import Base
from bluesky.ui import palette
import bluesky as bs


class AreaModel(QAbstractListModel, Base):
    polys: ActData[dict] = ActData(group='poly')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.polynames = list()

    def rowCount(self, parent=QModelIndex()):
        return len(self.polynames)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if index.isValid():
            # if role == Qt.ItemDataRole.DisplayRole:
            name = self.polynames[index.row()]
            data = self.polys.get(name)
            if data:
                return QVariant((name, self.polys[name]))
        return QVariant()

    def itemsRemoved(self, items):
        # remove items from the list
        pass

    @subscriber(topic='POLY')
    def on_poly_update(self, data):
        if ctx.action in (ctx.action.ActChange, ctx.action.Reset):
            # Notify the gui that the entire list needs to be updated
            self.beginResetModel()
            self.polynames = list(self.polys)
            self.endResetModel()

        elif ctx.action in (ctx.action.Append, ctx.action.Update):
            newpolys = ctx.action_content.get('polys', dict())
            newpolys = [name for name in newpolys if name not in self.polynames]
            if newpolys:
                self.beginInsertRows(QModelIndex(), len(self.polynames), len(self.polynames) + len(newpolys) - 1)
                self.polynames.extend(newpolys)
                self.endInsertRows()


class AreaList(QListView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setModel(AreaModel())
        self.setBackgroundRole(QPalette.ColorRole.NoRole)
        self.setAutoFillBackground(True)
        self.setStyleSheet('background-color: transparent')
        self.setItemDelegate(AreaItem())
        self.setMouseTracking(True)
        self.setSpacing(3)
        self.setSelectionMode(QListView.SelectionMode.SingleSelection)
        self.setSelectionBehavior(QListView.SelectionBehavior.SelectRows)
        self.selectionModel().selectionChanged.connect(
                self.on_change_selection
        )

    def on_change_selection(self) -> None:
        indexes = self.selectedIndexes()
        data = indexes[0].data()[1]
        coords = data['coordinates']
        lat = coords[::2]
        lon = coords[1::2]
        latrange = (min(lat), max(lat))
        lonrange = (min(lon), max(lon))
        pan = (0.5 * sum(latrange), 0.5 * sum(lonrange))
        flat_earth = cos(radians(pan[0]))
        latrange = latrange[1] - latrange[0]
        lonrange = (lonrange[1] - lonrange[0]) * flat_earth
        zoom = 1 / (max(latrange, lonrange))
        bs.stack.stack(f'PAN {pan[0]} {pan[1]};ZOOM {zoom}')


class AreaItem(QStyledItemDelegate):
    """ AreaItem is used to render area items in the area list """
    def paint(self, painter, option, index):
        # Determine dark mode
        p = app.instance().style().standardPalette()
        isdark = (p.color(p.ColorRole.Window).value() < p.color(p.ColorRole.WindowText).value())

        if isdark:
            bgcolor = p.color(p.ColorRole.Window).lighter(120)
            bghover = p.color(p.ColorRole.Window).lighter(140)
            txtcolor = Qt.GlobalColor.white
        else:
            bgcolor = QColor(250, 250, 250, 255)
            bghover = Qt.GlobalColor.lightGray
            txtcolor = Qt.GlobalColor.black

        # Get polygon name and data
        name, data = index.data()

        if option.state & QStyle.StateFlag.State_Selected:
            painter.setPen(Qt.GlobalColor.red)
        else:
            painter.setPen(txtcolor)
   
        if option.state & QStyle.StateFlag.State_MouseOver:
            painter.fillRect(option.rect, bghover)
        else:
            painter.fillRect(option.rect, bgcolor)
        r = QRect(option.rect)
        r.setWidth(10)

        color = QColor(*data.get('color', palette.polys), 255)
        painter.fillRect(r, color)
        # f"<b>ID:</b> {name}<br><b>Class:</b>{pdata['shape']}"
        items = [
            ('ID :', name),
            ('CLASS: ', data['shape'])
        ]
        r = QRect(option.rect)
        r.setLeft(r.left() + 15)
        for header, content in items:
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)
            painter.drawText(r, Qt.AlignmentFlag.AlignLeft, header)
            bb = painter.boundingRect(option.rect, Qt.AlignmentFlag.AlignLeft, header)
            font.setBold(False)
            painter.setFont(font)
            r.setLeft(r.left() + bb.width())
            painter.drawText(r, Qt.AlignmentFlag.AlignLeft, content)
            r.setLeft(option.rect.left() + 15)
            r.setTop(r.top() + bb.height() + 4)
        # else:
        #     QStyledItemDelegate.paint(self, painter, option, index)

    def sizeHint(self, option, index):
        """ Returns the size needed to display the item in a QSize object. """
        # index.data()
        return QSize(200, 45)




""" External control plugin for Machine Learning applications. """
# Import the global bluesky objects. Uncomment the ones you need
from bluesky import stack, net, sim, traf  #, settings, navdb, traf, sim, scr, tools


myclientrte = None

### Initialization function of your plugin. Do not change the name of this
### function, as it is the way BlueSky recognises this file as a plugin.
def init_plugin():

    # Addtional initilisation code

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'MLCONTROL',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim',

        # Update interval in seconds. By default, your plugin's update function(s)
        # are called every timestep of the simulation. If your plugin needs less
        # frequent updates provide an update interval.
        'update_interval': 1.0,

        'update':          update,

        # If your plugin has a state, you will probably need a reset function to
        # clear the state in between simulations.
        'reset':         reset
        }

    stackfunctions = {
        # The command name for your function
        'MLSTEP': [
            # A short usage string. This will be printed if you type HELP <name> in the BlueSky console
            'MLSTEP',

            # A list of the argument types your function accepts. For a description of this, see ...
            '',

            # The name of your function in this plugin
            mlstep,

            # a longer help text of your function.
            'Simulate one MLCONTROL time interval.']
    }

    # init_plugin() should always return these two dicts.
    return config, stackfunctions


### Periodic update functions that are called by the simulation. You can replace
### this by anything, so long as you communicate this in init_plugin

def update():
    data = dict(
        lat=traf.lat,
        lon=traf.lon,
        alt=traf.alt
    )
    net.send_event(b'MLSTATEREPLY', data, myclientrte)
    sim.hold()

def preupdate():
    pass

def reset():
    pass

def mlstep():
    global myclientrte
    myclientrte = stack.routetosender()
    sim.op()


from pathlib import Path
from bluesky.core import Entity
from bluesky.stack import commandgroup
from bluesky.stack.simstack import process, merge

class Importer(Entity):
    @commandgroup(name='IMPORT')
    @staticmethod
    def importcmd(fname:str='', *args):
        ''' Importer: file importer for 3rd party scenario data.
        
            Arguments:
            - fname: The filename of the file to import
            - args: Optional arguments to perform e.g., wildcard replacement
        '''
        if not fname:
            ret = Importer.importcmd.__doc__
            ret += '\nKnown importers:'
            for impl in Importer.derived().values():
                if impl is not Importer:
                    importer = impl.implinstance()
                    ret += f'\n{importer.filetype} (extensions: {", ".join(importer.extensions)})'
            return True, ret
        # Get the correct importer implementation
        ext = Path(fname).suffix.strip('.')
        for impl in Importer.derived().values():
            if impl is not Importer and ext in impl.implinstance().extensions:
                importer = impl.implinstance()
                break
        else:
            return False, f"No importer found for file with extension {ext}"

        return importer._load_and_process(fname)

    def __init__(self, filetype, extensions=None):
        super().__init__()
        self.filetype = filetype
        self.extensions = extensions or tuple()

    def _load_and_process(self, fname, *args):
        # Load the data, and merge with current scenario stack
        scentime, scencmd = self.load(fname, *args)
        # If there are no timestamps, process immediately
        if not scentime:
            process(zip(scencmd, len(scencmd) * [None]))
            return True, f'Successfully loaded {fname}'
        # If we have timestamps, merge the commands with the existing scenario stack
        merge(zip(scentime, scencmd))
        return True, f'Successfully loaded and merged {fname}'

    def load(fname, *args):
        ''' Importer file loading.
        
            Arguments:
            - fname: Name of the file to load.
            - args: Optional arguments to perform e.g., wildcard replacement

            Returns:
            - (scentime, scencmd): Tuple of two lists:
                - scentime: list of timestamps (in (fractions of) seconds). 
                  May be an empty list if commands are not timestamped.
                - scencmd: list of corresponding stack commands
        '''
        return ([], [])

    def __init_subclass__(cls):
        # First do base class initialisation
        super().__init_subclass__(replaceable=False, skipbase=True)
        # Then immediately instantiate
        cls._instance = cls()
        # Add subcommand for this specific importer
        Importer.importcmd.subcommand(cls._instance._load_and_process, name=cls._instance.filetype.upper())
